import os

# list all include directories
include_directories = [
    os.path.sep.join(x.split('/'))
    for x in ['extension/icu/include', 'extension/icu/third_party/icu/common', 'extension/icu/third_party/icu/i18n']
]
# source files
source_directories = [
    os.path.sep.join(x.split('/'))
    for x in ['.', 'third_party/icu/common', 'third_party/icu/i18n', 'third_party/icu/stubdata']
]
source_files = []
base_path = os.path.dirname(os.path.abspath(__file__))
for dir in source_directories:
    source_files += [
        os.path.join('extension', 'icu', dir, x) for x in os.listdir(os.path.join(base_path, dir)) if x.endswith('.cpp')
    ]



import os
version = '3.6.4'

# os.system(f'wget https://github.com/Mbed-TLS/mbedtls/archive/refs/tags/mbedtls-{version}.tar.gz')
# os.system(f'tar xvf mbedtls-{version}.tar.gz')

directories = ['include', 'library']
source_dir = f'mbedtls-mbedtls-{version}'
target_dir = '.'
extensions = ['.h', '.hpp', '.c', '.cpp']

class FileToCopy:
    def __init__(self, source, target):
        self.source_file = source
        self.target_file = target

def get_copy_list(source_dir, target_dir):
    result = []
    for file in os.listdir(source_dir):
        is_source_file = False
        for ext in extensions:
            if file.endswith(ext):
                is_source_file = True
        if not is_source_file:
            continue
        target_file_name = file
        if target_file_name.endswith('.c'):
            target_file_name = target_file_name[:-2] + '.cpp'
        source_file = os.path.join(source_dir, file)
        target_file = os.path.join(target_dir, target_file_name)
        if os.path.isdir(source_file):
            result += get_copy_list(source_file, target_file)
        if not os.path.isfile(target_file):
            continue
        # check if this is a dummy file
        with open(target_file, 'r') as f:
            text = f.read()
        if '// dummy file' in text:
            continue
        print(target_file)
        result.append(FileToCopy(source_file, target_file))
    return result


copy_list = []
for directory in directories:
    copy_list += get_copy_list(os.path.join(source_dir, directory), os.path.join(target_dir, directory))

for file in copy_list:
    os.system(f'cp {file.source_file} {file.target_file}')


import os

# list all include directories
include_directories = [
    os.path.sep.join(x.split('/'))
    for x in [
        'extension/parquet/include',
        'third_party/parquet',
        'third_party/thrift',
        'third_party/lz4',
        'third_party/brotli/include',
        'third_party/brotli/common',
        'third_party/brotli/dec',
        'third_party/brotli/enc',
        'third_party/snappy',
        'third_party/mbedtls',
        'third_party/mbedtls/include',
        'third_party/zstd/include',
    ]
]
prefix = os.path.join('extension', 'parquet')


def list_files_recursive(rootdir, suffix):
    file_list = []
    for root, _, files in os.walk(rootdir):
        file_list += [os.path.join(root, f) for f in files if f.endswith(suffix)]
    return file_list


source_files = list_files_recursive(prefix, '.cpp')

# parquet/thrift/snappy
source_files += [
    os.path.sep.join(x.split('/'))
    for x in [
        'third_party/parquet/parquet_types.cpp',
        'third_party/thrift/thrift/protocol/TProtocol.cpp',
        'third_party/thrift/thrift/transport/TTransportException.cpp',
        'third_party/thrift/thrift/transport/TBufferTransports.cpp',
        'third_party/snappy/snappy.cc',
        'third_party/snappy/snappy-sinksource.cc',
    ]
]
# lz4
source_files += [os.path.sep.join(x.split('/')) for x in ['third_party/lz4/lz4.cpp']]

# brotli
source_files += [
    os.path.sep.join(x.split('/'))
    for x in [
        'third_party/brotli/common/constants.cpp',
        'third_party/brotli/common/context.cpp',
        'third_party/brotli/common/dictionary.cpp',
        'third_party/brotli/common/platform.cpp',
        'third_party/brotli/common/shared_dictionary.cpp',
        'third_party/brotli/common/transform.cpp',
        'third_party/brotli/dec/bit_reader.cpp',
        'third_party/brotli/dec/decode.cpp',
        'third_party/brotli/dec/huffman.cpp',
        'third_party/brotli/dec/state.cpp',
        'third_party/brotli/enc/backward_references.cpp',
        'third_party/brotli/enc/backward_references_hq.cpp',
        'third_party/brotli/enc/bit_cost.cpp',
        'third_party/brotli/enc/block_splitter.cpp',
        'third_party/brotli/enc/brotli_bit_stream.cpp',
        'third_party/brotli/enc/cluster.cpp',
        'third_party/brotli/enc/command.cpp',
        'third_party/brotli/enc/compound_dictionary.cpp',
        'third_party/brotli/enc/compress_fragment.cpp',
        'third_party/brotli/enc/compress_fragment_two_pass.cpp',
        'third_party/brotli/enc/dictionary_hash.cpp',
        'third_party/brotli/enc/encode.cpp',
        'third_party/brotli/enc/encoder_dict.cpp',
        'third_party/brotli/enc/entropy_encode.cpp',
        'third_party/brotli/enc/fast_log.cpp',
        'third_party/brotli/enc/histogram.cpp',
        'third_party/brotli/enc/literal_cost.cpp',
        'third_party/brotli/enc/memory.cpp',
        'third_party/brotli/enc/metablock.cpp',
        'third_party/brotli/enc/static_dict.cpp',
        'third_party/brotli/enc/utf8_util.cpp',
    ]
]



import sys

contents = bytearray(sys.stdin.buffer.read())

# encode the data into the stubdata file
result_text = ",".join(map(str, contents))

new_contents = """
// This file is generated by scripts/inline-data.py, do not edit it manually //

#include "unicode/utypes.h"
#include "unicode/udata.h"
#include "unicode/uversion.h"

extern "C" U_EXPORT const unsigned char U_ICUDATA_ENTRY_POINT [] = {
	%s
};
""" % (
    result_text,
)

sys.stdout.write(new_contents)



# brotli uses a weird c templating mechanism using _inc.h files
# this does not play well with things like amalagamation
# this script inlines the variuos headers

import os
import re

for filename in os.listdir('.'):
    if not (filename.endswith('.cpp') or filename.endswith('.h')): 
        continue

    file_lines = open(filename, 'r').readlines()
    if '_inc.h' not in '\n'.join(file_lines):
        continue

    out = open (filename, 'w')

    for line in file_lines:
        if '#include' in line and '_inc.h' in line:
            match = re.search(r'#include\s+"(.+)".*', line).group(1)
            include = open(match, 'r').readlines();
            out.write(''.join(include))
            continue
        out.write(line)



import os
import pathlib
import typing as pytyping
from typing_extensions import Self

if pytyping.TYPE_CHECKING:
    import fsspec
    import numpy as np
    import polars
    import pandas
    import pyarrow.lib
    import torch as pytorch
    import tensorflow
    from collections.abc import Callable, Sequence, Mapping
    from duckdb import sqltypes, func

    # the field_ids argument to to_parquet and write_parquet has a recursive structure
    ParquetFieldIdsType = Mapping[str, pytyping.Union[int, "ParquetFieldIdsType"]]

__all__: list[str] = [
    "BinderException",
    "CSVLineTerminator",
    "CaseExpression",
    "CatalogException",
    "CoalesceOperator",
    "ColumnExpression",
    "ConnectionException",
    "ConstantExpression",
    "ConstraintException",
    "ConversionException",
    "DataError",
    "DatabaseError",
    "DefaultExpression",
    "DependencyException",
    "DuckDBPyConnection",
    "DuckDBPyRelation",
    "Error",
    "ExpectedResultType",
    "ExplainType",
    "Expression",
    "FatalException",
    "FunctionExpression",
    "HTTPException",
    "IOException",
    "IntegrityError",
    "InternalError",
    "InternalException",
    "InterruptException",
    "InvalidInputException",
    "InvalidTypeException",
    "LambdaExpression",
    "NotImplementedException",
    "NotSupportedError",
    "OperationalError",
    "OutOfMemoryException",
    "OutOfRangeException",
    "ParserException",
    "PermissionException",
    "ProgrammingError",
    "PythonExceptionHandling",
    "RenderMode",
    "SQLExpression",
    "SequenceException",
    "SerializationException",
    "StarExpression",
    "Statement",
    "StatementType",
    "SyntaxException",
    "TransactionException",
    "TypeMismatchException",
    "Warning",
    "aggregate",
    "alias",
    "apilevel",
    "append",
    "array_type",
    "arrow",
    "begin",
    "checkpoint",
    "close",
    "commit",
    "connect",
    "create_function",
    "cursor",
    "decimal_type",
    "default_connection",
    "description",
    "df",
    "distinct",
    "dtype",
    "duplicate",
    "enum_type",
    "execute",
    "executemany",
    "extract_statements",
    "fetch_arrow_table",
    "fetch_df",
    "fetch_df_chunk",
    "fetch_record_batch",
    "fetchall",
    "fetchdf",
    "fetchmany",
    "fetchnumpy",
    "fetchone",
    "filesystem_is_registered",
    "filter",
    "from_arrow",
    "from_csv_auto",
    "from_df",
    "from_parquet",
    "from_query",
    "get_table_names",
    "install_extension",
    "interrupt",
    "limit",
    "list_filesystems",
    "list_type",
    "load_extension",
    "map_type",
    "order",
    "paramstyle",
    "pl",
    "project",
    "query",
    "query_df",
    "query_progress",
    "read_csv",
    "read_json",
    "read_parquet",
    "register",
    "register_filesystem",
    "remove_function",
    "rollback",
    "row_type",
    "rowcount",
    "set_default_connection",
    "sql",
    "sqltype",
    "string_type",
    "struct_type",
    "table",
    "table_function",
    "tf",
    "threadsafety",
    "token_type",
    "tokenize",
    "torch",
    "type",
    "union_type",
    "unregister",
    "unregister_filesystem",
    "values",
    "view",
    "write_csv",
]

class BinderException(ProgrammingError): ...

class CSVLineTerminator:
    CARRIAGE_RETURN_LINE_FEED: pytyping.ClassVar[
        CSVLineTerminator
    ]  # value = <CSVLineTerminator.CARRIAGE_RETURN_LINE_FEED: 1>
    LINE_FEED: pytyping.ClassVar[CSVLineTerminator]  # value = <CSVLineTerminator.LINE_FEED: 0>
    __members__: pytyping.ClassVar[
        dict[str, CSVLineTerminator]
    ]  # value = {'LINE_FEED': <CSVLineTerminator.LINE_FEED: 0>, 'CARRIAGE_RETURN_LINE_FEED': <CSVLineTerminator.CARRIAGE_RETURN_LINE_FEED: 1>}  # noqa: E501
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class CatalogException(ProgrammingError): ...
class ConnectionException(OperationalError): ...
class ConstraintException(IntegrityError): ...
class ConversionException(DataError): ...
class DataError(DatabaseError): ...
class DatabaseError(Error): ...
class DependencyException(DatabaseError): ...

class DuckDBPyConnection:
    def __del__(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type: object, exc: object, traceback: object) -> None: ...
    def append(self, table_name: str, df: pandas.DataFrame, *, by_name: bool = False) -> DuckDBPyConnection: ...
    def array_type(self, type: sqltypes.DuckDBPyType, size: pytyping.SupportsInt) -> sqltypes.DuckDBPyType: ...
    def arrow(self, rows_per_batch: pytyping.SupportsInt = 1000000) -> pyarrow.lib.RecordBatchReader: ...
    def begin(self) -> DuckDBPyConnection: ...
    def checkpoint(self) -> DuckDBPyConnection: ...
    def close(self) -> None: ...
    def commit(self) -> DuckDBPyConnection: ...
    def create_function(
        self,
        name: str,
        function: Callable[..., pytyping.Any],
        parameters: list[sqltypes.DuckDBPyType] | None = None,
        return_type: sqltypes.DuckDBPyType | None = None,
        *,
        type: func.PythonUDFType = ...,
        null_handling: func.FunctionNullHandling = ...,
        exception_handling: PythonExceptionHandling = ...,
        side_effects: bool = False,
    ) -> DuckDBPyConnection: ...
    def cursor(self) -> DuckDBPyConnection: ...
    def decimal_type(self, width: pytyping.SupportsInt, scale: pytyping.SupportsInt) -> sqltypes.DuckDBPyType: ...
    def df(self, *, date_as_object: bool = False) -> pandas.DataFrame: ...
    def dtype(self, type_str: str) -> sqltypes.DuckDBPyType: ...
    def duplicate(self) -> DuckDBPyConnection: ...
    def enum_type(
        self, name: str, type: sqltypes.DuckDBPyType, values: list[pytyping.Any]
    ) -> sqltypes.DuckDBPyType: ...
    def execute(self, query: Statement | str, parameters: object = None) -> DuckDBPyConnection: ...
    def executemany(self, query: Statement | str, parameters: object = None) -> DuckDBPyConnection: ...
    def extract_statements(self, query: str) -> list[Statement]: ...
    def fetch_arrow_table(self, rows_per_batch: pytyping.SupportsInt = 1000000) -> pyarrow.lib.Table: ...
    def fetch_df(self, *, date_as_object: bool = False) -> pandas.DataFrame: ...
    def fetch_df_chunk(
        self, vectors_per_chunk: pytyping.SupportsInt = 1, *, date_as_object: bool = False
    ) -> pandas.DataFrame: ...
    def fetch_record_batch(self, rows_per_batch: pytyping.SupportsInt = 1000000) -> pyarrow.lib.RecordBatchReader: ...
    def fetchall(self) -> list[tuple[pytyping.Any, ...]]: ...
    def fetchdf(self, *, date_as_object: bool = False) -> pandas.DataFrame: ...
    def fetchmany(self, size: pytyping.SupportsInt = 1) -> list[tuple[pytyping.Any, ...]]: ...
    def fetchnumpy(self) -> dict[str, np.typing.NDArray[pytyping.Any] | pandas.Categorical]: ...
    def fetchone(self) -> tuple[pytyping.Any, ...] | None: ...
    def filesystem_is_registered(self, name: str) -> bool: ...
    def from_arrow(self, arrow_object: object) -> DuckDBPyRelation: ...
    def from_csv_auto(
        self,
        path_or_buffer: str | bytes | os.PathLike[str] | os.PathLike[bytes],
        header: bool | int | None = None,
        compression: str | None = None,
        sep: str | None = None,
        delimiter: str | None = None,
        files_to_sniff: int | None = None,
        comment: str | None = None,
        thousands: str | None = None,
        dtype: dict[str, str] | list[str] | None = None,
        na_values: str | list[str] | None = None,
        skiprows: int | None = None,
        quotechar: str | None = None,
        escapechar: str | None = None,
        encoding: str | None = None,
        parallel: bool | None = None,
        date_format: str | None = None,
        timestamp_format: str | None = None,
        sample_size: int | None = None,
        auto_detect: bool | int | None = None,
        all_varchar: bool | None = None,
        normalize_names: bool | None = None,
        null_padding: bool | None = None,
        names: list[str] | None = None,
        lineterminator: str | None = None,
        columns: dict[str, str] | None = None,
        auto_type_candidates: list[str] | None = None,
        max_line_size: int | None = None,
        ignore_errors: bool | None = None,
        store_rejects: bool | None = None,
        rejects_table: str | None = None,
        rejects_scan: str | None = None,
        rejects_limit: int | None = None,
        force_not_null: list[str] | None = None,
        buffer_size: int | None = None,
        decimal: str | None = None,
        allow_quoted_nulls: bool | None = None,
        filename: bool | str | None = None,
        hive_partitioning: bool | None = None,
        union_by_name: bool | None = None,
        hive_types: dict[str, str] | None = None,
        hive_types_autocast: bool | None = None,
        strict_mode: bool | None = None,
    ) -> DuckDBPyRelation: ...
    def from_df(self, df: pandas.DataFrame) -> DuckDBPyRelation: ...
    @pytyping.overload
    def from_parquet(
        self,
        file_glob: str,
        binary_as_string: bool = False,
        *,
        file_row_number: bool = False,
        filename: bool = False,
        hive_partitioning: bool = False,
        union_by_name: bool = False,
        compression: str | None = None,
    ) -> DuckDBPyRelation: ...
    @pytyping.overload
    def from_parquet(
        self,
        file_globs: Sequence[str],
        binary_as_string: bool = False,
        *,
        file_row_number: bool = False,
        filename: bool = False,
        hive_partitioning: bool = False,
        union_by_name: bool = False,
        compression: str | None = None,
    ) -> DuckDBPyRelation: ...
    def from_query(self, query: str, *, alias: str = "", params: object = None) -> DuckDBPyRelation: ...
    def get_table_names(self, query: str, *, qualified: bool = False) -> set[str]: ...
    def install_extension(
        self,
        extension: str,
        *,
        force_install: bool = False,
        repository: str | None = None,
        repository_url: str | None = None,
        version: str | None = None,
    ) -> None: ...
    def interrupt(self) -> None: ...
    def list_filesystems(self) -> list[str]: ...
    def list_type(self, type: sqltypes.DuckDBPyType) -> sqltypes.DuckDBPyType: ...
    def load_extension(self, extension: str) -> None: ...
    def map_type(self, key: sqltypes.DuckDBPyType, value: sqltypes.DuckDBPyType) -> sqltypes.DuckDBPyType: ...
    @pytyping.overload
    def pl(
        self, rows_per_batch: pytyping.SupportsInt = 1000000, *, lazy: pytyping.Literal[False] = ...
    ) -> polars.DataFrame: ...
    @pytyping.overload
    def pl(
        self, rows_per_batch: pytyping.SupportsInt = 1000000, *, lazy: pytyping.Literal[True]
    ) -> polars.LazyFrame: ...
    @pytyping.overload
    def pl(
        self, rows_per_batch: pytyping.SupportsInt = 1000000, *, lazy: bool = False
    ) -> pytyping.Union[polars.DataFrame, polars.LazyFrame]: ...
    def query(self, query: str, *, alias: str = "", params: object = None) -> DuckDBPyRelation: ...
    def query_progress(self) -> float: ...
    def read_csv(
        self,
        path_or_buffer: str | bytes | os.PathLike[str],
        header: bool | int | None = None,
        compression: str | None = None,
        sep: str | None = None,
        delimiter: str | None = None,
        files_to_sniff: int | None = None,
        comment: str | None = None,
        thousands: str | None = None,
        dtype: dict[str, str] | list[str] | None = None,
        na_values: str | list[str] | None = None,
        skiprows: int | None = None,
        quotechar: str | None = None,
        escapechar: str | None = None,
        encoding: str | None = None,
        parallel: bool | None = None,
        date_format: str | None = None,
        timestamp_format: str | None = None,
        sample_size: int | None = None,
        auto_detect: bool | int | None = None,
        all_varchar: bool | None = None,
        normalize_names: bool | None = None,
        null_padding: bool | None = None,
        names: list[str] | None = None,
        lineterminator: str | None = None,
        columns: dict[str, str] | None = None,
        auto_type_candidates: list[str] | None = None,
        max_line_size: int | None = None,
        ignore_errors: bool | None = None,
        store_rejects: bool | None = None,
        rejects_table: str | None = None,
        rejects_scan: str | None = None,
        rejects_limit: int | None = None,
        force_not_null: list[str] | None = None,
        buffer_size: int | None = None,
        decimal: str | None = None,
        allow_quoted_nulls: bool | None = None,
        filename: bool | str | None = None,
        hive_partitioning: bool | None = None,
        union_by_name: bool | None = None,
        hive_types: dict[str, str] | None = None,
        hive_types_autocast: bool | None = None,
        strict_mode: bool | None = None,
    ) -> DuckDBPyRelation: ...
    def read_json(
        self,
        path_or_buffer: str | bytes | os.PathLike[str],
        *,
        columns: dict[str, str] | None = None,
        sample_size: int | None = None,
        maximum_depth: int | None = None,
        records: str | None = None,
        format: str | None = None,
        date_format: str | None = None,
        timestamp_format: str | None = None,
        compression: str | None = None,
        maximum_object_size: int | None = None,
        ignore_errors: bool | None = None,
        convert_strings_to_integers: bool | None = None,
        field_appearance_threshold: float | None = None,
        map_inference_threshold: int | None = None,
        maximum_sample_files: int | None = None,
        filename: bool | str | None = None,
        hive_partitioning: bool | None = None,
        union_by_name: bool | None = None,
        hive_types: dict[str, str] | None = None,
        hive_types_autocast: bool | None = None,
    ) -> DuckDBPyRelation: ...
    @pytyping.overload
    def read_parquet(
        self,
        file_glob: str,
        binary_as_string: bool = False,
        *,
        file_row_number: bool = False,
        filename: bool = False,
        hive_partitioning: bool = False,
        union_by_name: bool = False,
        compression: str | None = None,
    ) -> DuckDBPyRelation: ...
    @pytyping.overload
    def read_parquet(
        self,
        file_globs: Sequence[str],
        binary_as_string: bool = False,
        *,
        file_row_number: bool = False,
        filename: bool = False,
        hive_partitioning: bool = False,
        union_by_name: bool = False,
        compression: pytyping.Any = None,
    ) -> DuckDBPyRelation: ...
    def register(self, view_name: str, python_object: object) -> DuckDBPyConnection: ...
    def register_filesystem(self, filesystem: fsspec.AbstractFileSystem) -> None: ...
    def remove_function(self, name: str) -> DuckDBPyConnection: ...
    def rollback(self) -> DuckDBPyConnection: ...
    def row_type(
        self, fields: dict[str, sqltypes.DuckDBPyType] | list[sqltypes.DuckDBPyType]
    ) -> sqltypes.DuckDBPyType: ...
    def sql(self, query: Statement | str, *, alias: str = "", params: object = None) -> DuckDBPyRelation: ...
    def sqltype(self, type_str: str) -> sqltypes.DuckDBPyType: ...
    def string_type(self, collation: str = "") -> sqltypes.DuckDBPyType: ...
    def struct_type(
        self, fields: dict[str, sqltypes.DuckDBPyType] | list[sqltypes.DuckDBPyType]
    ) -> sqltypes.DuckDBPyType: ...
    def table(self, table_name: str) -> DuckDBPyRelation: ...
    def table_function(self, name: str, parameters: object = None) -> DuckDBPyRelation: ...
    def tf(self) -> dict[str, tensorflow.Tensor]: ...
    def torch(self) -> dict[str, pytorch.Tensor]: ...
    def type(self, type_str: str) -> sqltypes.DuckDBPyType: ...
    def union_type(
        self, members: list[sqltypes.DuckDBPyType] | dict[str, sqltypes.DuckDBPyType]
    ) -> sqltypes.DuckDBPyType: ...
    def unregister(self, view_name: str) -> DuckDBPyConnection: ...
    def unregister_filesystem(self, name: str) -> None: ...
    def values(self, *args: list[pytyping.Any] | tuple[Expression, ...] | Expression) -> DuckDBPyRelation: ...
    def view(self, view_name: str) -> DuckDBPyRelation: ...
    @property
    def description(self) -> list[tuple[str, sqltypes.DuckDBPyType, None, None, None, None, None]]: ...
    @property
    def rowcount(self) -> int: ...

class DuckDBPyRelation:
    def __arrow_c_stream__(self, requested_schema: object | None = None) -> pytyping.Any: ...
    def __contains__(self, name: str) -> bool: ...
    def __getattr__(self, name: str) -> DuckDBPyRelation: ...
    def __getitem__(self, name: str) -> DuckDBPyRelation: ...
    def __len__(self) -> int: ...
    def aggregate(self, aggr_expr: Expression | str, group_expr: Expression | str = "") -> DuckDBPyRelation: ...
    def any_value(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def apply(
        self,
        function_name: str,
        function_aggr: str,
        group_expr: str = "",
        function_parameter: str = "",
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def arg_max(
        self, arg_column: str, value_column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def arg_min(
        self, arg_column: str, value_column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def arrow(self, batch_size: pytyping.SupportsInt = 1000000) -> pyarrow.lib.RecordBatchReader: ...
    def avg(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def bit_and(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def bit_or(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def bit_xor(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def bitstring_agg(
        self,
        column: str,
        min: int | None = None,
        max: int | None = None,
        groups: str = "",
        window_spec: str = "",
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def bool_and(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def bool_or(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def close(self) -> None: ...
    def count(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def create(self, table_name: str) -> None: ...
    def create_view(self, view_name: str, replace: bool = True) -> DuckDBPyRelation: ...
    def cross(self, other_rel: DuckDBPyRelation) -> DuckDBPyRelation: ...
    def cume_dist(self, window_spec: str, projected_columns: str = "") -> DuckDBPyRelation: ...
    def dense_rank(self, window_spec: str, projected_columns: str = "") -> DuckDBPyRelation: ...
    def describe(self) -> DuckDBPyRelation: ...
    def df(self, *, date_as_object: bool = False) -> pandas.DataFrame: ...
    def distinct(self) -> DuckDBPyRelation: ...
    def except_(self, other_rel: DuckDBPyRelation) -> DuckDBPyRelation: ...
    def execute(self) -> DuckDBPyRelation: ...
    def explain(self, type: ExplainType = ExplainType.STANDARD) -> str: ...
    def favg(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def fetch_arrow_reader(self, batch_size: pytyping.SupportsInt = 1000000) -> pyarrow.lib.RecordBatchReader: ...
    def fetch_arrow_table(self, batch_size: pytyping.SupportsInt = 1000000) -> pyarrow.lib.Table: ...
    def fetch_df_chunk(
        self, vectors_per_chunk: pytyping.SupportsInt = 1, *, date_as_object: bool = False
    ) -> pandas.DataFrame: ...
    def fetch_record_batch(self, rows_per_batch: pytyping.SupportsInt = 1000000) -> pyarrow.lib.RecordBatchReader: ...
    def fetchall(self) -> list[tuple[pytyping.Any, ...]]: ...
    def fetchdf(self, *, date_as_object: bool = False) -> pandas.DataFrame: ...
    def fetchmany(self, size: pytyping.SupportsInt = 1) -> list[tuple[pytyping.Any, ...]]: ...
    def fetchnumpy(self) -> dict[str, np.typing.NDArray[pytyping.Any] | pandas.Categorical]: ...
    def fetchone(self) -> tuple[pytyping.Any, ...] | None: ...
    def filter(self, filter_expr: Expression | str) -> DuckDBPyRelation: ...
    def first(self, column: str, groups: str = "", projected_columns: str = "") -> DuckDBPyRelation: ...
    def first_value(self, column: str, window_spec: str = "", projected_columns: str = "") -> DuckDBPyRelation: ...
    def fsum(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def geomean(self, column: str, groups: str = "", projected_columns: str = "") -> DuckDBPyRelation: ...
    def histogram(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def insert(self, values: pytyping.List[object]) -> None: ...
    def insert_into(self, table_name: str) -> None: ...
    def intersect(self, other_rel: DuckDBPyRelation) -> DuckDBPyRelation: ...
    def join(
        self, other_rel: DuckDBPyRelation, condition: Expression | str, how: str = "inner"
    ) -> DuckDBPyRelation: ...
    def lag(
        self,
        column: str,
        window_spec: str,
        offset: pytyping.SupportsInt = 1,
        default_value: str = "NULL",
        ignore_nulls: bool = False,
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def last(self, column: str, groups: str = "", projected_columns: str = "") -> DuckDBPyRelation: ...
    def last_value(self, column: str, window_spec: str = "", projected_columns: str = "") -> DuckDBPyRelation: ...
    def lead(
        self,
        column: str,
        window_spec: str,
        offset: pytyping.SupportsInt = 1,
        default_value: str = "NULL",
        ignore_nulls: bool = False,
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def limit(self, n: pytyping.SupportsInt, offset: pytyping.SupportsInt = 0) -> DuckDBPyRelation: ...
    def list(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def map(
        self, map_function: Callable[..., pytyping.Any], *, schema: dict[str, sqltypes.DuckDBPyType] | None = None
    ) -> DuckDBPyRelation: ...
    def max(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def mean(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def median(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def min(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def mode(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def n_tile(
        self, window_spec: str, num_buckets: pytyping.SupportsInt, projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def nth_value(
        self,
        column: str,
        window_spec: str,
        offset: pytyping.SupportsInt,
        ignore_nulls: bool = False,
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def order(self, order_expr: str) -> DuckDBPyRelation: ...
    def percent_rank(self, window_spec: str, projected_columns: str = "") -> DuckDBPyRelation: ...
    @pytyping.overload
    def pl(
        self, batch_size: pytyping.SupportsInt = 1000000, *, lazy: pytyping.Literal[False] = ...
    ) -> polars.DataFrame: ...
    @pytyping.overload
    def pl(self, batch_size: pytyping.SupportsInt = 1000000, *, lazy: pytyping.Literal[True]) -> polars.LazyFrame: ...
    @pytyping.overload
    def pl(
        self, batch_size: pytyping.SupportsInt = 1000000, *, lazy: bool = False
    ) -> pytyping.Union[polars.DataFrame, polars.LazyFrame]: ...
    def product(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def project(self, *args: str | Expression, groups: str = "") -> DuckDBPyRelation: ...
    def quantile(
        self,
        column: str,
        q: float | pytyping.List[float] = 0.5,
        groups: str = "",
        window_spec: str = "",
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def quantile_cont(
        self,
        column: str,
        q: float | pytyping.List[float] = 0.5,
        groups: str = "",
        window_spec: str = "",
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def quantile_disc(
        self,
        column: str,
        q: float | pytyping.List[float] = 0.5,
        groups: str = "",
        window_spec: str = "",
        projected_columns: str = "",
    ) -> DuckDBPyRelation: ...
    def query(self, virtual_table_name: str, sql_query: str) -> DuckDBPyRelation: ...
    def rank(self, window_spec: str, projected_columns: str = "") -> DuckDBPyRelation: ...
    def rank_dense(self, window_spec: str, projected_columns: str = "") -> DuckDBPyRelation: ...
    def record_batch(self, batch_size: pytyping.SupportsInt = 1000000) -> pyarrow.RecordBatchReader: ...
    def row_number(self, window_spec: str, projected_columns: str = "") -> DuckDBPyRelation: ...
    def select(self, *args: str | Expression, groups: str = "") -> DuckDBPyRelation: ...
    def select_dtypes(self, types: pytyping.List[sqltypes.DuckDBPyType | str]) -> DuckDBPyRelation: ...
    def select_types(self, types: pytyping.List[sqltypes.DuckDBPyType | str]) -> DuckDBPyRelation: ...
    def set_alias(self, alias: str) -> DuckDBPyRelation: ...
    def show(
        self,
        *,
        max_width: pytyping.SupportsInt | None = None,
        max_rows: pytyping.SupportsInt | None = None,
        max_col_width: pytyping.SupportsInt | None = None,
        null_value: str | None = None,
        render_mode: RenderMode | None = None,
    ) -> None: ...
    def sort(self, *args: Expression) -> DuckDBPyRelation: ...
    def sql_query(self) -> str: ...
    def std(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def stddev(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def stddev_pop(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def stddev_samp(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def string_agg(
        self, column: str, sep: str = ",", groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def sum(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def tf(self) -> dict[str, tensorflow.Tensor]: ...
    def to_arrow_table(self, batch_size: pytyping.SupportsInt = 1000000) -> pyarrow.lib.Table: ...
    def to_csv(
        self,
        file_name: str,
        *,
        sep: str | None = None,
        na_rep: str | None = None,
        header: bool | None = None,
        quotechar: str | None = None,
        escapechar: str | None = None,
        date_format: str | None = None,
        timestamp_format: str | None = None,
        quoting: str | int | None = None,
        encoding: str | None = None,
        compression: str | None = None,
        overwrite: bool | None = None,
        per_thread_output: bool | None = None,
        use_tmp_file: bool | None = None,
        partition_by: pytyping.List[str] | None = None,
        write_partition_columns: bool | None = None,
    ) -> None: ...
    def to_df(self, *, date_as_object: bool = False) -> pandas.DataFrame: ...
    def to_parquet(
        self,
        file_name: str,
        *,
        compression: str | None = None,
        field_ids: ParquetFieldIdsType | pytyping.Literal["auto"] | None = None,
        row_group_size_bytes: int | str | None = None,
        row_group_size: int | None = None,
        overwrite: bool | None = None,
        per_thread_output: bool | None = None,
        use_tmp_file: bool | None = None,
        partition_by: pytyping.List[str] | None = None,
        write_partition_columns: bool | None = None,
        append: bool | None = None,
        filename_pattern: str | None = None,
        file_size_bytes: str | int | None = None,
    ) -> None: ...
    def to_table(self, table_name: str) -> None: ...
    def to_view(self, view_name: str, replace: bool = True) -> DuckDBPyRelation: ...
    def torch(self) -> dict[str, pytorch.Tensor]: ...
    def union(self, union_rel: DuckDBPyRelation) -> DuckDBPyRelation: ...
    def unique(self, unique_aggr: str) -> DuckDBPyRelation: ...
    def update(self, set: Expression | str, *, condition: Expression | str | None = None) -> None: ...
    def value_counts(self, column: str, groups: str = "") -> DuckDBPyRelation: ...
    def var(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def var_pop(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def var_samp(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def variance(
        self, column: str, groups: str = "", window_spec: str = "", projected_columns: str = ""
    ) -> DuckDBPyRelation: ...
    def write_csv(
        self,
        file_name: str,
        sep: str | None = None,
        na_rep: str | None = None,
        header: bool | None = None,
        quotechar: str | None = None,
        escapechar: str | None = None,
        date_format: str | None = None,
        timestamp_format: str | None = None,
        quoting: str | int | None = None,
        encoding: str | None = None,
        compression: str | None = None,
        overwrite: bool | None = None,
        per_thread_output: bool | None = None,
        use_tmp_file: bool | None = None,
        partition_by: pytyping.List[str] | None = None,
        write_partition_columns: bool | None = None,
    ) -> None: ...
    def write_parquet(
        self,
        file_name: str,
        compression: str | None = None,
        field_ids: ParquetFieldIdsType | pytyping.Literal["auto"] | None = None,
        row_group_size_bytes: str | int | None = None,
        row_group_size: int | None = None,
        overwrite: bool | None = None,
        per_thread_output: bool | None = None,
        use_tmp_file: bool | None = None,
        partition_by: pytyping.List[str] | None = None,
        write_partition_columns: bool | None = None,
        append: bool | None = None,
        filename_pattern: str | None = None,
        file_size_bytes: str | int | None = None,
    ) -> None: ...
    @property
    def alias(self) -> str: ...
    @property
    def columns(self) -> pytyping.List[str]: ...
    @property
    def description(self) -> pytyping.List[tuple[str, sqltypes.DuckDBPyType, None, None, None, None, None]]: ...
    @property
    def dtypes(self) -> pytyping.List[str]: ...
    @property
    def shape(self) -> tuple[int, int]: ...
    @property
    def type(self) -> str: ...
    @property
    def types(self) -> pytyping.List[sqltypes.DuckDBPyType]: ...

class Error(Exception): ...

class ExpectedResultType:
    CHANGED_ROWS: pytyping.ClassVar[ExpectedResultType]  # value = <ExpectedResultType.CHANGED_ROWS: 1>
    NOTHING: pytyping.ClassVar[ExpectedResultType]  # value = <ExpectedResultType.NOTHING: 2>
    QUERY_RESULT: pytyping.ClassVar[ExpectedResultType]  # value = <ExpectedResultType.QUERY_RESULT: 0>
    __members__: pytyping.ClassVar[
        dict[str, ExpectedResultType]
    ]  # value = {'QUERY_RESULT': <ExpectedResultType.QUERY_RESULT: 0>, 'CHANGED_ROWS': <ExpectedResultType.CHANGED_ROWS: 1>, 'NOTHING': <ExpectedResultType.NOTHING: 2>}  # noqa: E501
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class ExplainType:
    ANALYZE: pytyping.ClassVar[ExplainType]  # value = <ExplainType.ANALYZE: 1>
    STANDARD: pytyping.ClassVar[ExplainType]  # value = <ExplainType.STANDARD: 0>
    __members__: pytyping.ClassVar[
        dict[str, ExplainType]
    ]  # value = {'STANDARD': <ExplainType.STANDARD: 0>, 'ANALYZE': <ExplainType.ANALYZE: 1>}
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class Expression:
    def __add__(self, other: Expression) -> Expression: ...
    def __and__(self, other: Expression) -> Expression: ...
    def __div__(self, other: Expression) -> Expression: ...
    def __eq__(self, other: Expression) -> Expression: ...  # type: ignore[override]
    def __floordiv__(self, other: Expression) -> Expression: ...
    def __ge__(self, other: Expression) -> Expression: ...
    def __gt__(self, other: Expression) -> Expression: ...
    @pytyping.overload
    def __init__(self, arg0: str) -> None: ...
    @pytyping.overload
    def __init__(self, arg0: pytyping.Any) -> None: ...
    def __invert__(self) -> Expression: ...
    def __le__(self, other: Expression) -> Expression: ...
    def __lt__(self, other: Expression) -> Expression: ...
    def __mod__(self, other: Expression) -> Expression: ...
    def __mul__(self, other: Expression) -> Expression: ...
    def __ne__(self, other: Expression) -> Expression: ...  # type: ignore[override]
    def __neg__(self) -> Expression: ...
    def __or__(self, other: Expression) -> Expression: ...
    def __pow__(self, other: Expression) -> Expression: ...
    def __radd__(self, other: Expression) -> Expression: ...
    def __rand__(self, other: Expression) -> Expression: ...
    def __rdiv__(self, other: Expression) -> Expression: ...
    def __rfloordiv__(self, other: Expression) -> Expression: ...
    def __rmod__(self, other: Expression) -> Expression: ...
    def __rmul__(self, other: Expression) -> Expression: ...
    def __ror__(self, other: Expression) -> Expression: ...
    def __rpow__(self, other: Expression) -> Expression: ...
    def __rsub__(self, other: Expression) -> Expression: ...
    def __rtruediv__(self, other: Expression) -> Expression: ...
    def __sub__(self, other: Expression) -> Expression: ...
    def __truediv__(self, other: Expression) -> Expression: ...
    def alias(self, name: str) -> Expression: ...
    def asc(self) -> Expression: ...
    def between(self, lower: Expression, upper: Expression) -> Expression: ...
    def cast(self, type: sqltypes.DuckDBPyType) -> Expression: ...
    def collate(self, collation: str) -> Expression: ...
    def desc(self) -> Expression: ...
    def get_name(self) -> str: ...
    def isin(self, *args: Expression) -> Expression: ...
    def isnotin(self, *args: Expression) -> Expression: ...
    def isnotnull(self) -> Expression: ...
    def isnull(self) -> Expression: ...
    def nulls_first(self) -> Expression: ...
    def nulls_last(self) -> Expression: ...
    def otherwise(self, value: Expression) -> Expression: ...
    def show(self) -> None: ...
    def when(self, condition: Expression, value: Expression) -> Expression: ...

class FatalException(DatabaseError): ...

class HTTPException(IOException):
    status_code: int
    body: str
    reason: str
    headers: dict[str, str]

class IOException(OperationalError): ...
class IntegrityError(DatabaseError): ...
class InternalError(DatabaseError): ...
class InternalException(InternalError): ...
class InterruptException(DatabaseError): ...
class InvalidInputException(ProgrammingError): ...
class InvalidTypeException(ProgrammingError): ...
class NotImplementedException(NotSupportedError): ...
class NotSupportedError(DatabaseError): ...
class OperationalError(DatabaseError): ...
class OutOfMemoryException(OperationalError): ...
class OutOfRangeException(DataError): ...
class ParserException(ProgrammingError): ...
class PermissionException(DatabaseError): ...
class ProgrammingError(DatabaseError): ...

class PythonExceptionHandling:
    DEFAULT: pytyping.ClassVar[PythonExceptionHandling]  # value = <PythonExceptionHandling.DEFAULT: 0>
    RETURN_NULL: pytyping.ClassVar[PythonExceptionHandling]  # value = <PythonExceptionHandling.RETURN_NULL: 1>
    __members__: pytyping.ClassVar[
        dict[str, PythonExceptionHandling]
    ]  # value = {'DEFAULT': <PythonExceptionHandling.DEFAULT: 0>, 'RETURN_NULL': <PythonExceptionHandling.RETURN_NULL: 1>}  # noqa: E501
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class RenderMode:
    COLUMNS: pytyping.ClassVar[RenderMode]  # value = <RenderMode.COLUMNS: 1>
    ROWS: pytyping.ClassVar[RenderMode]  # value = <RenderMode.ROWS: 0>
    __members__: pytyping.ClassVar[
        dict[str, RenderMode]
    ]  # value = {'ROWS': <RenderMode.ROWS: 0>, 'COLUMNS': <RenderMode.COLUMNS: 1>}
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class SequenceException(DatabaseError): ...
class SerializationException(OperationalError): ...

class Statement:
    @property
    def expected_result_type(self) -> list[StatementType]: ...
    @property
    def named_parameters(self) -> set[str]: ...
    @property
    def query(self) -> str: ...
    @property
    def type(self) -> StatementType: ...

class StatementType:
    ALTER_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.ALTER_STATEMENT: 8>
    ANALYZE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.ANALYZE_STATEMENT: 11>
    ATTACH_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.ATTACH_STATEMENT: 25>
    CALL_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.CALL_STATEMENT: 19>
    COPY_DATABASE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.COPY_DATABASE_STATEMENT: 28>
    COPY_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.COPY_STATEMENT: 10>
    CREATE_FUNC_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.CREATE_FUNC_STATEMENT: 13>
    CREATE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.CREATE_STATEMENT: 4>
    DELETE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.DELETE_STATEMENT: 5>
    DETACH_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.DETACH_STATEMENT: 26>
    DROP_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.DROP_STATEMENT: 15>
    EXECUTE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.EXECUTE_STATEMENT: 7>
    EXPLAIN_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.EXPLAIN_STATEMENT: 14>
    EXPORT_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.EXPORT_STATEMENT: 16>
    EXTENSION_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.EXTENSION_STATEMENT: 23>
    INSERT_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.INSERT_STATEMENT: 2>
    INVALID_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.INVALID_STATEMENT: 0>
    LOAD_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.LOAD_STATEMENT: 21>
    LOGICAL_PLAN_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.LOGICAL_PLAN_STATEMENT: 24>
    MERGE_INTO_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.MERGE_INTO_STATEMENT: 30>
    MULTI_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.MULTI_STATEMENT: 27>
    PRAGMA_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.PRAGMA_STATEMENT: 17>
    PREPARE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.PREPARE_STATEMENT: 6>
    RELATION_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.RELATION_STATEMENT: 22>
    SELECT_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.SELECT_STATEMENT: 1>
    SET_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.SET_STATEMENT: 20>
    TRANSACTION_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.TRANSACTION_STATEMENT: 9>
    UPDATE_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.UPDATE_STATEMENT: 3>
    VACUUM_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.VACUUM_STATEMENT: 18>
    VARIABLE_SET_STATEMENT: pytyping.ClassVar[StatementType]  # value = <StatementType.VARIABLE_SET_STATEMENT: 12>
    __members__: pytyping.ClassVar[
        dict[str, StatementType]
    ]  # value = {'INVALID_STATEMENT': <StatementType.INVALID_STATEMENT: 0>, 'SELECT_STATEMENT': <StatementType.SELECT_STATEMENT: 1>, 'INSERT_STATEMENT': <StatementType.INSERT_STATEMENT: 2>, 'UPDATE_STATEMENT': <StatementType.UPDATE_STATEMENT: 3>, 'CREATE_STATEMENT': <StatementType.CREATE_STATEMENT: 4>, 'DELETE_STATEMENT': <StatementType.DELETE_STATEMENT: 5>, 'PREPARE_STATEMENT': <StatementType.PREPARE_STATEMENT: 6>, 'EXECUTE_STATEMENT': <StatementType.EXECUTE_STATEMENT: 7>, 'ALTER_STATEMENT': <StatementType.ALTER_STATEMENT: 8>, 'TRANSACTION_STATEMENT': <StatementType.TRANSACTION_STATEMENT: 9>, 'COPY_STATEMENT': <StatementType.COPY_STATEMENT: 10>, 'ANALYZE_STATEMENT': <StatementType.ANALYZE_STATEMENT: 11>, 'VARIABLE_SET_STATEMENT': <StatementType.VARIABLE_SET_STATEMENT: 12>, 'CREATE_FUNC_STATEMENT': <StatementType.CREATE_FUNC_STATEMENT: 13>, 'EXPLAIN_STATEMENT': <StatementType.EXPLAIN_STATEMENT: 14>, 'DROP_STATEMENT': <StatementType.DROP_STATEMENT: 15>, 'EXPORT_STATEMENT': <StatementType.EXPORT_STATEMENT: 16>, 'PRAGMA_STATEMENT': <StatementType.PRAGMA_STATEMENT: 17>, 'VACUUM_STATEMENT': <StatementType.VACUUM_STATEMENT: 18>, 'CALL_STATEMENT': <StatementType.CALL_STATEMENT: 19>, 'SET_STATEMENT': <StatementType.SET_STATEMENT: 20>, 'LOAD_STATEMENT': <StatementType.LOAD_STATEMENT: 21>, 'RELATION_STATEMENT': <StatementType.RELATION_STATEMENT: 22>, 'EXTENSION_STATEMENT': <StatementType.EXTENSION_STATEMENT: 23>, 'LOGICAL_PLAN_STATEMENT': <StatementType.LOGICAL_PLAN_STATEMENT: 24>, 'ATTACH_STATEMENT': <StatementType.ATTACH_STATEMENT: 25>, 'DETACH_STATEMENT': <StatementType.DETACH_STATEMENT: 26>, 'MULTI_STATEMENT': <StatementType.MULTI_STATEMENT: 27>, 'COPY_DATABASE_STATEMENT': <StatementType.COPY_DATABASE_STATEMENT: 28>, 'MERGE_INTO_STATEMENT': <StatementType.MERGE_INTO_STATEMENT: 30>}  # noqa: E501
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class SyntaxException(ProgrammingError): ...
class TransactionException(OperationalError): ...
class TypeMismatchException(DataError): ...
class Warning(Exception): ...

class token_type:
    __members__: pytyping.ClassVar[
        dict[str, token_type]
    ]  # value = {'identifier': <token_type.identifier: 0>, 'numeric_const': <token_type.numeric_const: 1>, 'string_const': <token_type.string_const: 2>, 'operator': <token_type.operator: 3>, 'keyword': <token_type.keyword: 4>, 'comment': <token_type.comment: 5>}  # noqa: E501
    comment: pytyping.ClassVar[token_type]  # value = <token_type.comment: 5>
    identifier: pytyping.ClassVar[token_type]  # value = <token_type.identifier: 0>
    keyword: pytyping.ClassVar[token_type]  # value = <token_type.keyword: 4>
    numeric_const: pytyping.ClassVar[token_type]  # value = <token_type.numeric_const: 1>
    operator: pytyping.ClassVar[token_type]  # value = <token_type.operator: 3>
    string_const: pytyping.ClassVar[token_type]  # value = <token_type.string_const: 2>
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: pytyping.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: pytyping.SupportsInt) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

def CaseExpression(condition: Expression, value: Expression) -> Expression: ...
def CoalesceOperator(*args: Expression) -> Expression: ...
def ColumnExpression(*args: str) -> Expression: ...
def ConstantExpression(value: pytyping.Any) -> Expression: ...
def DefaultExpression() -> Expression: ...
def FunctionExpression(function_name: str, *args: Expression) -> Expression: ...
def LambdaExpression(lhs: pytyping.Any, rhs: Expression) -> Expression: ...
def SQLExpression(expression: str) -> Expression: ...
def StarExpression(*, exclude: pytyping.Any = None) -> Expression: ...
def aggregate(
    df: pandas.DataFrame,
    aggr_expr: Expression | list[Expression] | str | list[str],
    group_expr: str = "",
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def alias(df: pandas.DataFrame, alias: str, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
def append(
    table_name: str, df: pandas.DataFrame, *, by_name: bool = False, connection: DuckDBPyConnection | None = None
) -> DuckDBPyConnection: ...
def array_type(
    type: sqltypes.DuckDBPyType, size: pytyping.SupportsInt, *, connection: DuckDBPyConnection | None = None
) -> sqltypes.DuckDBPyType: ...
@pytyping.overload
def arrow(
    rows_per_batch: pytyping.SupportsInt = 1000000, *, connection: DuckDBPyConnection | None = None
) -> pyarrow.lib.RecordBatchReader: ...
@pytyping.overload
def arrow(arrow_object: pytyping.Any, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
def begin(*, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def checkpoint(*, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def close(*, connection: DuckDBPyConnection | None = None) -> None: ...
def commit(*, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def connect(
    database: str | pathlib.Path = ":memory:",
    read_only: bool = False,
    config: dict[str, str | bool | int | float | list[str]] | None = None,
) -> DuckDBPyConnection: ...
def create_function(
    name: str,
    function: Callable[..., pytyping.Any],
    parameters: list[sqltypes.DuckDBPyType] | None = None,
    return_type: sqltypes.DuckDBPyType | None = None,
    *,
    type: func.PythonUDFType = ...,
    null_handling: func.FunctionNullHandling = ...,
    exception_handling: PythonExceptionHandling = ...,
    side_effects: bool = False,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyConnection: ...
def cursor(*, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def decimal_type(
    width: pytyping.SupportsInt, scale: pytyping.SupportsInt, *, connection: DuckDBPyConnection | None = None
) -> sqltypes.DuckDBPyType: ...
def default_connection() -> DuckDBPyConnection: ...
def description(
    *, connection: DuckDBPyConnection | None = None
) -> list[tuple[str, sqltypes.DuckDBPyType, None, None, None, None, None]] | None: ...
@pytyping.overload
def df(*, date_as_object: bool = False, connection: DuckDBPyConnection | None = None) -> pandas.DataFrame: ...
@pytyping.overload
def df(df: pandas.DataFrame, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
def distinct(df: pandas.DataFrame, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
def dtype(type_str: str, *, connection: DuckDBPyConnection | None = None) -> sqltypes.DuckDBPyType: ...
def duplicate(*, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def enum_type(
    name: str,
    type: sqltypes.DuckDBPyType,
    values: list[pytyping.Any],
    *,
    connection: DuckDBPyConnection | None = None,
) -> sqltypes.DuckDBPyType: ...
def execute(
    query: Statement | str,
    parameters: object = None,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyConnection: ...
def executemany(
    query: Statement | str,
    parameters: object = None,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyConnection: ...
def extract_statements(query: str, *, connection: DuckDBPyConnection | None = None) -> list[Statement]: ...
def fetch_arrow_table(
    rows_per_batch: pytyping.SupportsInt = 1000000, *, connection: DuckDBPyConnection | None = None
) -> pyarrow.lib.Table: ...
def fetch_df(*, date_as_object: bool = False, connection: DuckDBPyConnection | None = None) -> pandas.DataFrame: ...
def fetch_df_chunk(
    vectors_per_chunk: pytyping.SupportsInt = 1,
    *,
    date_as_object: bool = False,
    connection: DuckDBPyConnection | None = None,
) -> pandas.DataFrame: ...
def fetch_record_batch(
    rows_per_batch: pytyping.SupportsInt = 1000000, *, connection: DuckDBPyConnection | None = None
) -> pyarrow.lib.RecordBatchReader: ...
def fetchall(*, connection: DuckDBPyConnection | None = None) -> list[tuple[pytyping.Any, ...]]: ...
def fetchdf(*, date_as_object: bool = False, connection: DuckDBPyConnection | None = None) -> pandas.DataFrame: ...
def fetchmany(
    size: pytyping.SupportsInt = 1, *, connection: DuckDBPyConnection | None = None
) -> list[tuple[pytyping.Any, ...]]: ...
def fetchnumpy(
    *, connection: DuckDBPyConnection | None = None
) -> dict[str, np.typing.NDArray[pytyping.Any] | pandas.Categorical]: ...
def fetchone(*, connection: DuckDBPyConnection | None = None) -> tuple[pytyping.Any, ...] | None: ...
def filesystem_is_registered(name: str, *, connection: DuckDBPyConnection | None = None) -> bool: ...
def filter(
    df: pandas.DataFrame,
    filter_expr: Expression | str,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def from_arrow(
    arrow_object: object,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def from_csv_auto(
    path_or_buffer: str | bytes | os.PathLike[str],
    header: bool | int | None = None,
    compression: str | None = None,
    sep: str | None = None,
    delimiter: str | None = None,
    files_to_sniff: int | None = None,
    comment: str | None = None,
    thousands: str | None = None,
    dtype: dict[str, str] | list[str] | None = None,
    na_values: str | list[str] | None = None,
    skiprows: int | None = None,
    quotechar: str | None = None,
    escapechar: str | None = None,
    encoding: str | None = None,
    parallel: bool | None = None,
    date_format: str | None = None,
    timestamp_format: str | None = None,
    sample_size: int | None = None,
    auto_detect: bool | int | None = None,
    all_varchar: bool | None = None,
    normalize_names: bool | None = None,
    null_padding: bool | None = None,
    names: list[str] | None = None,
    lineterminator: str | None = None,
    columns: dict[str, str] | None = None,
    auto_type_candidates: list[str] | None = None,
    max_line_size: int | None = None,
    ignore_errors: bool | None = None,
    store_rejects: bool | None = None,
    rejects_table: str | None = None,
    rejects_scan: str | None = None,
    rejects_limit: int | None = None,
    force_not_null: list[str] | None = None,
    buffer_size: int | None = None,
    decimal: str | None = None,
    allow_quoted_nulls: bool | None = None,
    filename: bool | str | None = None,
    hive_partitioning: bool | None = None,
    union_by_name: bool | None = None,
    hive_types: dict[str, str] | None = None,
    hive_types_autocast: bool | None = None,
    strict_mode: bool | None = None,
) -> DuckDBPyRelation: ...
def from_df(df: pandas.DataFrame, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
@pytyping.overload
def from_parquet(
    file_glob: str,
    binary_as_string: bool = False,
    *,
    file_row_number: bool = False,
    filename: bool = False,
    hive_partitioning: bool = False,
    union_by_name: bool = False,
    compression: str | None = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
@pytyping.overload
def from_parquet(
    file_globs: Sequence[str],
    binary_as_string: bool = False,
    *,
    file_row_number: bool = False,
    filename: bool = False,
    hive_partitioning: bool = False,
    union_by_name: bool = False,
    compression: pytyping.Any = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def from_query(
    query: Statement | str,
    *,
    alias: str = "",
    params: object = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def get_table_names(
    query: str, *, qualified: bool = False, connection: DuckDBPyConnection | None = None
) -> set[str]: ...
def install_extension(
    extension: str,
    *,
    force_install: bool = False,
    repository: str | None = None,
    repository_url: str | None = None,
    version: str | None = None,
    connection: DuckDBPyConnection | None = None,
) -> None: ...
def interrupt(*, connection: DuckDBPyConnection | None = None) -> None: ...
def limit(
    df: pandas.DataFrame,
    n: pytyping.SupportsInt,
    offset: pytyping.SupportsInt = 0,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def list_filesystems(*, connection: DuckDBPyConnection | None = None) -> list[str]: ...
def list_type(
    type: sqltypes.DuckDBPyType, *, connection: DuckDBPyConnection | None = None
) -> sqltypes.DuckDBPyType: ...
def load_extension(extension: str, *, connection: DuckDBPyConnection | None = None) -> None: ...
def map_type(
    key: sqltypes.DuckDBPyType,
    value: sqltypes.DuckDBPyType,
    *,
    connection: DuckDBPyConnection | None = None,
) -> sqltypes.DuckDBPyType: ...
def order(
    df: pandas.DataFrame, order_expr: str, *, connection: DuckDBPyConnection | None = None
) -> DuckDBPyRelation: ...
@pytyping.overload
def pl(
    rows_per_batch: pytyping.SupportsInt = 1000000,
    *,
    lazy: pytyping.Literal[False] = ...,
    connection: DuckDBPyConnection | None = None,
) -> polars.DataFrame: ...
@pytyping.overload
def pl(
    rows_per_batch: pytyping.SupportsInt = 1000000,
    *,
    lazy: pytyping.Literal[True],
    connection: DuckDBPyConnection | None = None,
) -> polars.LazyFrame: ...
@pytyping.overload
def pl(
    rows_per_batch: pytyping.SupportsInt = 1000000,
    *,
    lazy: bool = False,
    connection: DuckDBPyConnection | None = None,
) -> pytyping.Union[polars.DataFrame, polars.LazyFrame]: ...
def project(
    df: pandas.DataFrame, *args: str | Expression, groups: str = "", connection: DuckDBPyConnection | None = None
) -> DuckDBPyRelation: ...
def query(
    query: Statement | str,
    *,
    alias: str = "",
    params: object = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def query_df(
    df: pandas.DataFrame,
    virtual_table_name: str,
    sql_query: str,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def query_progress(*, connection: DuckDBPyConnection | None = None) -> float: ...
def read_csv(
    path_or_buffer: str | bytes | os.PathLike[str],
    header: bool | int | None = None,
    compression: str | None = None,
    sep: str | None = None,
    delimiter: str | None = None,
    files_to_sniff: int | None = None,
    comment: str | None = None,
    thousands: str | None = None,
    dtype: dict[str, str] | list[str] | None = None,
    na_values: str | list[str] | None = None,
    skiprows: int | None = None,
    quotechar: str | None = None,
    escapechar: str | None = None,
    encoding: str | None = None,
    parallel: bool | None = None,
    date_format: str | None = None,
    timestamp_format: str | None = None,
    sample_size: int | None = None,
    auto_detect: bool | int | None = None,
    all_varchar: bool | None = None,
    normalize_names: bool | None = None,
    null_padding: bool | None = None,
    names: list[str] | None = None,
    lineterminator: str | None = None,
    columns: dict[str, str] | None = None,
    auto_type_candidates: list[str] | None = None,
    max_line_size: int | None = None,
    ignore_errors: bool | None = None,
    store_rejects: bool | None = None,
    rejects_table: str | None = None,
    rejects_scan: str | None = None,
    rejects_limit: int | None = None,
    force_not_null: list[str] | None = None,
    buffer_size: int | None = None,
    decimal: str | None = None,
    allow_quoted_nulls: bool | None = None,
    filename: bool | str | None = None,
    hive_partitioning: bool | None = None,
    union_by_name: bool | None = None,
    hive_types: dict[str, str] | None = None,
    hive_types_autocast: bool | None = None,
    strict_mode: bool | None = None,
) -> DuckDBPyRelation: ...
def read_json(
    path_or_buffer: str | bytes | os.PathLike[str],
    *,
    columns: dict[str, str] | None = None,
    sample_size: int | None = None,
    maximum_depth: int | None = None,
    records: str | None = None,
    format: str | None = None,
    date_format: str | None = None,
    timestamp_format: str | None = None,
    compression: str | None = None,
    maximum_object_size: int | None = None,
    ignore_errors: bool | None = None,
    convert_strings_to_integers: bool | None = None,
    field_appearance_threshold: float | None = None,
    map_inference_threshold: int | None = None,
    maximum_sample_files: int | None = None,
    filename: bool | str | None = None,
    hive_partitioning: bool | None = None,
    union_by_name: bool | None = None,
    hive_types: dict[str, str] | None = None,
    hive_types_autocast: bool | None = None,
) -> DuckDBPyRelation: ...
@pytyping.overload
def read_parquet(
    file_glob: str,
    binary_as_string: bool = False,
    *,
    file_row_number: bool = False,
    filename: bool = False,
    hive_partitioning: bool = False,
    union_by_name: bool = False,
    compression: str | None = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
@pytyping.overload
def read_parquet(
    file_globs: Sequence[str],
    binary_as_string: bool = False,
    *,
    file_row_number: bool = False,
    filename: bool = False,
    hive_partitioning: bool = False,
    union_by_name: bool = False,
    compression: pytyping.Any = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def register(
    view_name: str,
    python_object: object,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyConnection: ...
def register_filesystem(
    filesystem: fsspec.AbstractFileSystem, *, connection: DuckDBPyConnection | None = None
) -> None: ...
def remove_function(name: str, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def rollback(*, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def row_type(
    fields: dict[str, sqltypes.DuckDBPyType] | list[sqltypes.DuckDBPyType],
    *,
    connection: DuckDBPyConnection | None = None,
) -> sqltypes.DuckDBPyType: ...
def rowcount(*, connection: DuckDBPyConnection | None = None) -> int: ...
def set_default_connection(connection: DuckDBPyConnection) -> None: ...
def sql(
    query: Statement | str,
    *,
    alias: str = "",
    params: object = None,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def sqltype(type_str: str, *, connection: DuckDBPyConnection | None = None) -> sqltypes.DuckDBPyType: ...
def string_type(collation: str = "", *, connection: DuckDBPyConnection | None = None) -> sqltypes.DuckDBPyType: ...
def struct_type(
    fields: dict[str, sqltypes.DuckDBPyType] | list[sqltypes.DuckDBPyType],
    *,
    connection: DuckDBPyConnection | None = None,
) -> sqltypes.DuckDBPyType: ...
def table(table_name: str, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
def table_function(
    name: str,
    parameters: object = None,
    *,
    connection: DuckDBPyConnection | None = None,
) -> DuckDBPyRelation: ...
def tf(*, connection: DuckDBPyConnection | None = None) -> dict[str, tensorflow.Tensor]: ...
def tokenize(query: str) -> list[tuple[int, token_type]]: ...
def torch(*, connection: DuckDBPyConnection | None = None) -> dict[str, pytorch.Tensor]: ...
def type(type_str: str, *, connection: DuckDBPyConnection | None = None) -> sqltypes.DuckDBPyType: ...
def union_type(
    members: dict[str, sqltypes.DuckDBPyType] | list[sqltypes.DuckDBPyType],
    *,
    connection: DuckDBPyConnection | None = None,
) -> sqltypes.DuckDBPyType: ...
def unregister(view_name: str, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyConnection: ...
def unregister_filesystem(name: str, *, connection: DuckDBPyConnection | None = None) -> None: ...
def values(
    *args: list[pytyping.Any] | tuple[Expression, ...] | Expression, connection: DuckDBPyConnection | None = None
) -> DuckDBPyRelation: ...
def view(view_name: str, *, connection: DuckDBPyConnection | None = None) -> DuckDBPyRelation: ...
def write_csv(
    df: pandas.DataFrame,
    filename: str,
    *,
    sep: str | None = None,
    na_rep: str | None = None,
    header: bool | None = None,
    quotechar: str | None = None,
    escapechar: str | None = None,
    date_format: str | None = None,
    timestamp_format: str | None = None,
    quoting: str | int | None = None,
    encoding: str | None = None,
    compression: str | None = None,
    overwrite: bool | None = None,
    per_thread_output: bool | None = None,
    use_tmp_file: bool | None = None,
    partition_by: list[str] | None = None,
    write_partition_columns: bool | None = None,
) -> None: ...

__formatted_python_version__: str
__git_revision__: str
__interactive__: bool
__jupyter__: bool
__standard_vector_size__: int
__version__: str
_clean_default_connection: pytyping.Any  # value = <capsule object>
apilevel: str
paramstyle: str
threadsafety: int



"""DuckDB function constants and types. DEPRECATED: please use `duckdb.func` instead."""

import warnings

from duckdb.func import ARROW, DEFAULT, NATIVE, SPECIAL, FunctionNullHandling, PythonUDFType

__all__ = ["ARROW", "DEFAULT", "NATIVE", "SPECIAL", "FunctionNullHandling", "PythonUDFType"]

warnings.warn(
    "`duckdb.functional` is deprecated and will be removed in a future version. Please use `duckdb.func` instead.",
    DeprecationWarning,
    stacklevel=2,
)



import os

# list all include directories
include_directories = [
    os.path.sep.join(x.split('/')) for x in ['extension/tpch/include', 'extension/tpch/dbgen/include']
]
# source files
source_files = [
    os.path.sep.join(x.split('/'))
    for x in [
        'extension/tpch/tpch_extension.cpp',
        'extension/tpch/dbgen/bm_utils.cpp',
        'extension/tpch/dbgen/build.cpp',
        'extension/tpch/dbgen/dbgen.cpp',
        'extension/tpch/dbgen/dbgen_gunk.cpp',
        'extension/tpch/dbgen/permute.cpp',
        'extension/tpch/dbgen/rnd.cpp',
        'extension/tpch/dbgen/rng64.cpp',
        'extension/tpch/dbgen/speed_seed.cpp',
        'extension/tpch/dbgen/text.cpp',
    ]
]



import argparse  # noqa: D100
import json
import re
import webbrowser
from functools import reduce
from pathlib import Path

qgraph_css = """
.styled-table {
	border-collapse: collapse;
	margin: 25px 0;
	font-size: 0.9em;
	font-family: sans-serif;
	min-width: 400px;
	box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
}
.styled-table thead tr {
	background-color: #009879;
	color: #ffffff;
	text-align: left;
}
.styled-table th,
.styled-table td {
	padding: 12px 15px;
}
.styled-table tbody tr {
	border-bottom: 1px solid #dddddd;
}

.styled-table tbody tr:nth-of-type(even) {
	background-color: #f3f3f3;
}

.styled-table tbody tr:last-of-type {
	border-bottom: 2px solid #009879;
}

.node-body {
	font-size:15px;
}
.tf-nc {
	position: relative;
	width: 180px;
	text-align: center;
	background-color: #fff100;
}
.custom-tooltip {
  position: relative;
  display: inline-block;
}

.tooltip-text {
  visibility: hidden;
  background-color: #333;
  color: #fff;
  text-align: center;
  padding: 0px;
  border-radius: 1px;

  /* Positioning */
  position: absolute;
  z-index: 1;
  bottom: 100%;
  left: 50%;
  transform: translateX(-50%);
  margin-bottom: 8px;

  /* Tooltip Arrow */
  width: 400px;
}

.custom-tooltip:hover .tooltip-text {
  visibility: visible;
}
"""


class NodeTiming:  # noqa: D101
    def __init__(self, phase: str, time: float) -> None:  # noqa: D107
        self.phase = phase
        self.time = time
        # percentage is determined later.
        self.percentage = 0

    def calculate_percentage(self, total_time: float) -> None:  # noqa: D102
        self.percentage = self.time / total_time

    def combine_timing(self, r: "NodeTiming") -> "NodeTiming":  # noqa: D102
        # TODO: can only add timings for same-phase nodes  # noqa: TD002, TD003
        total_time = self.time + r.time
        return NodeTiming(self.phase, total_time)


class AllTimings:  # noqa: D101
    def __init__(self) -> None:  # noqa: D107
        self.phase_to_timings = {}

    def add_node_timing(self, node_timing: NodeTiming) -> None:  # noqa: D102
        if node_timing.phase in self.phase_to_timings:
            self.phase_to_timings[node_timing.phase].append(node_timing)
        else:
            self.phase_to_timings[node_timing.phase] = [node_timing]

    def get_phase_timings(self, phase: str) -> list[NodeTiming]:  # noqa: D102
        return self.phase_to_timings[phase]

    def get_summary_phase_timings(self, phase: str) -> NodeTiming:  # noqa: D102
        return reduce(NodeTiming.combine_timing, self.phase_to_timings[phase])

    def get_phases(self) -> list[NodeTiming]:  # noqa: D102
        phases = list(self.phase_to_timings.keys())
        phases.sort(key=lambda x: (self.get_summary_phase_timings(x)).time)
        phases.reverse()
        return phases

    def get_sum_of_all_timings(self) -> float:  # noqa: D102
        total_timing_sum = 0
        for phase in self.phase_to_timings:
            total_timing_sum += self.get_summary_phase_timings(phase).time
        return total_timing_sum


def open_utf8(fpath: str, flags: str) -> object:  # noqa: D103
    return Path(fpath).open(mode=flags, encoding="utf8")


def get_child_timings(top_node: object, query_timings: object) -> str:  # noqa: D103
    node_timing = NodeTiming(top_node["operator_type"], float(top_node["operator_timing"]))
    query_timings.add_node_timing(node_timing)
    for child in top_node["children"]:
        get_child_timings(child, query_timings)


def get_pink_shade_hex(fraction: float) -> str:  # noqa: D103
    fraction = max(0, min(1, fraction))

    # Define the RGB values for very light pink (almost white) and dark pink
    light_pink = (255, 250, 250)  # Very light pink
    dark_pink = (255, 20, 147)  # Dark pink

    # Calculate the RGB values for the given fraction
    r = int(light_pink[0] + (dark_pink[0] - light_pink[0]) * fraction)
    g = int(light_pink[1] + (dark_pink[1] - light_pink[1]) * fraction)
    b = int(light_pink[2] + (dark_pink[2] - light_pink[2]) * fraction)

    # Return as hexadecimal color code
    return f"#{r:02x}{g:02x}{b:02x}"


def get_node_body(name: str, result: str, cpu_time: float, card: int, est: int, width: int, extra_info: str) -> str:  # noqa: D103
    node_style = f"background-color: {get_pink_shade_hex(float(result) / cpu_time)};"

    body = f'<span class="tf-nc custom-tooltip" style="{node_style}">'
    body += '<div class="node-body">'
    new_name = "BRIDGE" if (name == "INVALID") else name.replace("_", " ")
    formatted_num = f"{float(result):.4f}"
    body += f"<p><b>{new_name}</b> </p><p>time: {formatted_num} seconds</p>"
    body += f'<span class="tooltip-text"> {extra_info} </span>'
    if width > 0:
        body += f"<p>cardinality: {card}</p>"
        body += f"<p>estimate: {est}</p>"
        body += f"<p>width: {width} bytes</p>"
    # TODO: Expand on timing. Usually available from a detailed profiling  # noqa: TD002, TD003
    body += "</div>"
    body += "</span>"
    return body


def generate_tree_recursive(json_graph: object, cpu_time: float) -> str:  # noqa: D103
    node_prefix_html = "<li>"
    node_suffix_html = "</li>"

    extra_info = ""
    estimate = 0
    for key in json_graph["extra_info"]:
        value = json_graph["extra_info"][key]
        if key == "Estimated Cardinality":
            estimate = int(value)
        else:
            extra_info += f"{key}: {value} <br>"
    cardinality = json_graph["operator_cardinality"]
    width = int(json_graph["result_set_size"] / max(1, cardinality))

    # get rid of some typically long names
    extra_info = re.sub(r"__internal_\s*", "__", extra_info)
    extra_info = re.sub(r"compress_integral\s*", "compress", extra_info)

    node_body = get_node_body(
        json_graph["operator_type"],
        json_graph["operator_timing"],
        cpu_time,
        cardinality,
        estimate,
        width,
        re.sub(r",\s*", ", ", extra_info),
    )

    children_html = ""
    if len(json_graph["children"]) >= 1:
        children_html += "<ul>"
        for child in json_graph["children"]:
            children_html += generate_tree_recursive(child, cpu_time)
        children_html += "</ul>"
    return node_prefix_html + node_body + children_html + node_suffix_html


# For generating the table in the top left.
def generate_timing_html(graph_json: object, query_timings: object) -> object:  # noqa: D103
    json_graph = json.loads(graph_json)
    gather_timing_information(json_graph, query_timings)
    total_time = float(json_graph.get("operator_timing") or json_graph.get("latency"))
    table_head = """
	<table class=\"styled-table\">
		<thead>
			<tr>
				<th>Phase</th>
				<th>Time</th>
				<th>Percentage</th>
			</tr>
		</thead>"""

    table_body = "<tbody>"
    table_end = "</tbody></table>"

    execution_time = query_timings.get_sum_of_all_timings()

    all_phases = query_timings.get_phases()
    query_timings.add_node_timing(NodeTiming("TOTAL TIME", total_time))
    query_timings.add_node_timing(NodeTiming("Execution Time", execution_time))
    all_phases = ["TOTAL TIME", "Execution Time", *all_phases]
    for phase in all_phases:
        summarized_phase = query_timings.get_summary_phase_timings(phase)
        summarized_phase.calculate_percentage(total_time)
        phase_column = f"<b>{phase}</b>" if phase == "TOTAL TIME" or phase == "Execution Time" else phase
        table_body += f"""
	<tr>
			<td>{phase_column}</td>
            <td>{summarized_phase.time}</td>
            <td>{str(summarized_phase.percentage * 100)[:6]}%</td>
    </tr>
"""
    table_body += table_end
    return table_head + table_body


def generate_tree_html(graph_json: object) -> str:  # noqa: D103
    json_graph = json.loads(graph_json)
    cpu_time = float(json_graph["cpu_time"])
    tree_prefix = '<div class="tf-tree tf-gap-sm"> \n <ul>'
    tree_suffix = "</ul> </div>"
    # first level of json is general overview
    # TODO: make sure json output first level always has only 1 level  # noqa: TD002, TD003
    tree_body = generate_tree_recursive(json_graph["children"][0], cpu_time)
    return tree_prefix + tree_body + tree_suffix


def generate_ipython(json_input: str) -> str:  # noqa: D103
    from IPython.core.display import HTML

    html_output = generate_html(json_input, False)  # noqa: F821

    return HTML(
        ('\n	${CSS}\n	${LIBRARIES}\n	<div class="chart" id="query-profile"></div>\n	${CHART_SCRIPT}\n	')
        .replace("${CSS}", html_output["css"])
        .replace("${CHART_SCRIPT}", html_output["chart_script"])
        .replace("${LIBRARIES}", html_output["libraries"])
    )


def generate_style_html(graph_json: str, include_meta_info: bool) -> None:  # noqa: D103, FBT001
    treeflex_css = '<link rel="stylesheet" href="https://unpkg.com/treeflex/dist/css/treeflex.css">\n'
    css = "<style>\n"
    css += qgraph_css + "\n"
    css += "</style>\n"
    return {"treeflex_css": treeflex_css, "duckdb_css": css, "libraries": "", "chart_script": ""}


def gather_timing_information(json: str, query_timings: object) -> None:  # noqa: D103
    # add up all of the times
    # measure each time as a percentage of the total time.
    # then you can return a list of [phase, time, percentage]
    get_child_timings(json["children"][0], query_timings)


def translate_json_to_html(input_file: str, output_file: str) -> None:  # noqa: D103
    query_timings = AllTimings()
    with open_utf8(input_file, "r") as f:
        text = f.read()

    html_output = generate_style_html(text, True)
    timing_table = generate_timing_html(text, query_timings)
    tree_output = generate_tree_html(text)

    # finally create and write the html
    with open_utf8(output_file, "w+") as f:
        html = """<!DOCTYPE html>
<html>
	<head>
	<meta charset="utf-8">
	<meta name="viewport" content="width=device-width">
	<title>Query Profile Graph for Query</title>
	${TREEFLEX_CSS}
	<style>
		${DUCKDB_CSS}
	</style>
</head>
<body>
	<div id="meta-info"></div>
	<div class="chart" id="query-profile">
		${TIMING_TABLE}
	</div>
	${TREE}
</body>
</html>
"""
        html = html.replace("${TREEFLEX_CSS}", html_output["treeflex_css"])
        html = html.replace("${DUCKDB_CSS}", html_output["duckdb_css"])
        html = html.replace("${TIMING_TABLE}", timing_table)
        html = html.replace("${TREE}", tree_output)
        f.write(html)


def main() -> None:  # noqa: D103
    parser = argparse.ArgumentParser(
        prog="Query Graph Generator",
        description="""Given a json profile output, generate a html file showing the query graph and
        timings of operators""",
    )
    parser.add_argument("profile_input", help="profile input in json")
    parser.add_argument("--out", required=False, default=False)
    parser.add_argument("--open", required=False, action="store_true", default=True)
    args = parser.parse_args()

    input = args.profile_input
    output = args.out
    if not args.out:
        if ".json" in input:
            output = input.replace(".json", ".html")
        else:
            print("please provide profile output in json")
            exit(1)
    else:
        if ".html" in args.out:
            output = args.out
        else:
            print("please provide valid .html file for output name")
            exit(1)

    open_output = args.open

    translate_json_to_html(input, output)

    if open_output:
        webbrowser.open(f"file://{Path(output).resolve()}", new=2)


if __name__ == "__main__":
    main()



import os
import argparse
from pathlib import Path

parser = argparse.ArgumentParser(description='Inline the auto-complete PEG grammar files')
parser.add_argument(
    '--print', action='store_true', help='Print the grammar instead of writing to a file', default=False
)
parser.add_argument(
    '--grammar-file',
    action='store_true',
    help='Write the grammar to a .gram file instead of a C++ header',
    default=False,
)

args = parser.parse_args()

autocomplete_dir = Path(__file__).parent
statements_dir = os.path.join(autocomplete_dir, 'grammar', 'statements')
keywords_dir = os.path.join(autocomplete_dir, 'grammar', 'keywords')
target_file = os.path.join(autocomplete_dir, 'include', 'inlined_grammar.hpp')

contents = ""

# Maps filenames to string categories
FILENAME_TO_CATEGORY = {
    "reserved_keyword.list": "RESERVED_KEYWORD",
    "unreserved_keyword.list": "UNRESERVED_KEYWORD",
    "column_name_keyword.list": "COL_NAME_KEYWORD",
    "func_name_keyword.list": "TYPE_FUNC_NAME_KEYWORD",
    "type_name_keyword.list": "TYPE_FUNC_NAME_KEYWORD",
}

# Maps category names to their C++ map variable names
CPP_MAP_NAMES = {
    "RESERVED_KEYWORD": "reserved_keyword_map",
    "UNRESERVED_KEYWORD": "unreserved_keyword_map",
    "COL_NAME_KEYWORD": "colname_keyword_map",
    "TYPE_FUNC_NAME_KEYWORD": "typefunc_keyword_map",
}

# Use a dictionary of sets to collect keywords for each category, preventing duplicates
keyword_sets = {category: set() for category in CPP_MAP_NAMES.keys()}

# --- Validation and Loading (largely unchanged) ---
# For validation during the loading phase
reserved_set = set()
unreserved_set = set()


def load_keywords(filepath):
    with open(filepath, "r") as f:
        return [line.strip().lower() for line in f if line.strip()]


for filename in os.listdir(keywords_dir):
    if filename not in FILENAME_TO_CATEGORY:
        continue

    category = FILENAME_TO_CATEGORY[filename]
    keywords = load_keywords(os.path.join(keywords_dir, filename))

    for kw in keywords:
        # Validation logic remains the same to enforce rules
        if category == "RESERVED_KEYWORD":
            if kw in reserved_set or kw in unreserved_set:
                print(f"Keyword '{kw}' has conflicting RESERVED/UNRESERVED categories")
                exit(1)
            reserved_set.add(kw)
        elif category == "UNRESERVED_KEYWORD":
            if kw in reserved_set or kw in unreserved_set:
                print(f"Keyword '{kw}' has conflicting RESERVED/UNRESERVED categories")
                exit(1)
            unreserved_set.add(kw)

        # Add the keyword to the appropriate set
        keyword_sets[category].add(kw)

# --- C++ Code Generation ---
output_path = os.path.join(autocomplete_dir, "keyword_map.cpp")
with open(output_path, "w") as f:
    f.write("/* THIS FILE WAS AUTOMATICALLY GENERATED BY inline_grammar.py */\n")
    f.write("#include \"keyword_helper.hpp\"\n\n")
    f.write("namespace duckdb {\n")
    f.write("void KeywordHelper::InitializeKeywordMaps() { // Renamed for clarity\n")
    f.write("\tif (initialized) {\n\t\treturn;\n\t};\n")
    f.write("\tinitialized = true;\n\n")

    # Get the total number of categories to handle the last item differently
    num_categories = len(keyword_sets)

    # Iterate through each category and generate code for each map
    for i, (category, keywords) in enumerate(keyword_sets.items()):
        cpp_map_name = CPP_MAP_NAMES[category]
        f.write(f"\t// Populating {cpp_map_name}\n")
        # Sort keywords for deterministic output
        for kw in sorted(list(keywords)):
            # Populate the C++ set with insert
            f.write(f'\t{cpp_map_name}.insert("{kw}");\n')

        # Add a newline for all but the last block
        if i < num_categories - 1:
            f.write("\n")
    f.write("}\n")
    f.write("} // namespace duckdb\n")

print(f"Successfully generated {output_path}")


def filename_to_upper_camel(file):
    name, _ = os.path.splitext(file)  # column_name_keywords
    parts = name.split('_')  # ['column', 'name', 'keywords']
    return ''.join(p.capitalize() for p in parts)


for file in os.listdir(keywords_dir):
    if not file.endswith('.list'):
        continue
    rule_name = filename_to_upper_camel(file)
    rule = f"{rule_name} <- "
    with open(os.path.join(keywords_dir, file), 'r') as f:
        lines = [f"'{line.strip()}'i" for line in f if line.strip()]
        rule += " /\n".join(lines) + "\n"
    contents += rule

for file in os.listdir(statements_dir):
    if not file.endswith('.gram'):
        raise Exception(f"File {file} does not end with .gram")
    with open(os.path.join(statements_dir, file), 'r') as f:
        contents += f.read() + "\n"

if args.print:
    print(contents)
    exit(0)

if args.grammar_file:
    grammar_file = target_file.replace('.hpp', '.gram')
    with open(grammar_file, 'w+') as f:
        f.write(contents)
    exit(0)


def get_grammar_bytes(contents, add_null_terminator=True):
    result_text = ""
    for line in contents.split('\n'):
        if len(line) == 0:
            continue
        result_text += "\t\"" + line.replace('\\', '\\\\').replace('"', '\\"') + "\\n\"\n"
    return result_text


with open(target_file, 'w+') as f:
    f.write(
        '''/* THIS FILE WAS AUTOMATICALLY GENERATED BY inline_grammar.py */
#pragma once

namespace duckdb {

const char INLINED_PEG_GRAMMAR[] = {
'''
        + get_grammar_bytes(contents)
        + '''
};

} // namespace duckdb
'''
    )



import os

# list all include directories
include_directories = [os.path.sep.join(x.split('/')) for x in ['extension/json/include']]


# source files
def list_files_recursive(rootdir, suffix):
    file_list = []
    for root, _, files in os.walk(rootdir):
        file_list += [os.path.join(root, f) for f in files if f.endswith(suffix)]
    return file_list


prefix = os.path.join('extension', 'json')
source_files = list_files_recursive(prefix, '.cpp')



"""DuckDB PEP 517 and PEP 660 build backend.

This module wraps the scikit-build-core build backend because:
1. We need to be able to determine the version of the DuckDB submodule while building
   a source distribution, so that we can pass it in when building a wheel. The backend
   tries to figure out which duckdb version will be included in the sdist and saves
   the output
2. We want to use a custom version scheme with setuptools-scm, and PEP 621 provides no
   way to specify local code as a build-backend plugin. However, PEP 517 allows us to
   put our own build backend on the python path with the `build.backend-path` key. The
   side effect is that our version scheme is also on the path during the build.

Also see https://peps.python.org/pep-0517/#in-tree-build-backends.
"""

import subprocess
import sys
from pathlib import Path
from typing import Optional, Union

from scikit_build_core.build import (
    build_editable,
    get_requires_for_build_editable,
    get_requires_for_build_sdist,
    get_requires_for_build_wheel,
    prepare_metadata_for_build_editable,
    prepare_metadata_for_build_wheel,
)
from scikit_build_core.build import (
    build_sdist as skbuild_build_sdist,
)
from scikit_build_core.build import (
    build_wheel as skbuild_build_wheel,
)

from duckdb_packaging._versioning import get_git_describe, pep440_to_git_tag, strip_post_from_version
from duckdb_packaging.setuptools_scm_version import MAIN_BRANCH_VERSIONING, forced_version_from_env

_DUCKDB_VERSION_FILENAME = "duckdb_version.txt"
_LOGGING_FORMAT = "[duckdb_pytooling.build_backend] {}"
_SKBUILD_CMAKE_OVERRIDE_GIT_DESCRIBE = "cmake.define.OVERRIDE_GIT_DESCRIBE"
# The below will check whether we should set a specific version in our build, and if so, set the version
_FORCED_PEP440_VERSION = forced_version_from_env()


def _log(msg: str) -> None:
    """Log a message with build backend prefix.

    Args:
        msg: The message to log.
    """
    print(_LOGGING_FORMAT.format(msg), flush=True, file=sys.stderr)


def _in_git_repository() -> bool:
    """Check if the current directory is inside a git repository.

    Returns:
        True if .git directory exists, False otherwise.
    """
    return Path(".git").exists()


def _in_sdist() -> bool:
    """Check if the current directory is inside a git repository.

    Returns:
        True if the duckdb version file exists and PKG-INFO exists, False otherwise.
    """
    return _version_file_path().exists() and Path("PKG-INFO").exists()


def _duckdb_submodule_path() -> Path:
    """Verify that the duckdb submodule is checked out and usable and return its path."""
    if not _in_git_repository():
        msg = "Not in a git repository, no duckdb submodule present"
        raise RuntimeError(msg)
    # search the duckdb submodule
    gitmodules_path = Path(".gitmodules")
    modules = {}
    with gitmodules_path.open("r") as f:
        cur_module_path = None
        cur_module_reponame = None
        for line in f:
            if line.strip().startswith("[submodule"):
                if cur_module_reponame is not None and cur_module_path is not None:
                    modules[cur_module_reponame] = cur_module_path
                    cur_module_reponame = None
                    cur_module_path = None
            elif line.strip().startswith("path"):
                cur_module_path = line.split("=")[-1].strip()
            elif line.strip().startswith("url"):
                basename = Path(line.split("=")[-1].strip()).name
                cur_module_reponame = basename[:-4] if basename.endswith(".git") else basename
        if cur_module_reponame is not None and cur_module_path is not None:
            modules[cur_module_reponame] = cur_module_path

    if "duckdb" not in modules:
        msg = "DuckDB submodule missing"
        raise RuntimeError(msg)

    duckdb_path = modules["duckdb"]
    # now check that the submodule is usable
    proc = subprocess.Popen(["git", "submodule", "status", duckdb_path], stdout=subprocess.PIPE)
    status, _ = proc.communicate()
    status = status.decode("ascii", "replace")
    for line in status.splitlines():
        if line.startswith("-"):
            msg = f"Duckdb submodule not initialized: {line}"
            raise RuntimeError(msg)
        if line.startswith("U"):
            msg = f"Duckdb submodule has merge conflicts: {line}"
            raise RuntimeError(msg)
        if line.startswith("+"):
            _log(f"WARNING: Duckdb submodule not clean: {line}")
    # all good
    return Path(duckdb_path)


def _version_file_path() -> Path:
    package_dir = Path(__file__).parent
    return package_dir / _DUCKDB_VERSION_FILENAME


def _write_duckdb_long_version(long_version: str) -> None:
    """Write the given version string to a file in the same directory as this module."""
    _version_file_path().write_text(long_version, encoding="utf-8")


def _read_duckdb_long_version() -> str:
    """Read the given version string from a file in the same directory as this module."""
    return _version_file_path().read_text(encoding="utf-8").strip()


def _skbuild_config_add(key: str, value: Union[list, str], config_settings: dict[str, Union[list[str], str]]) -> None:
    """Add or modify a configuration setting for scikit-build-core.

    This function handles adding values to scikit-build-core configuration settings,
    supporting both string and list types with appropriate merging behavior.

    Args:
        key: The configuration key to set (will be prefixed with 'skbuild.' if needed).
        value: The value to add (string or list).
        config_settings: The configuration dictionary to modify.

    Raises:
        RuntimeError: If this would overwrite an existing value, or on type mismatches.
        AssertionError: If config_settings is None.

    Behavior Rules:
        - String value + list setting: value is appended to the list
        - String value + string setting: existing value is overridden
        - List value + list setting: existing list is extended
        - List value + string setting: raises RuntimeError

    Note:
        scikit-build-core's preference logic for config sources still applies,
        considering env vars, config_settings and pyproject in that order,
        without merging between those sources.
    """
    assert config_settings is not None, "config_settings must not be None"
    store_key = key if key in config_settings else "skbuild." + key
    key_exists = store_key in config_settings
    key_exists_as_str = key_exists and isinstance(config_settings[store_key], str)
    key_exists_as_list = key_exists and isinstance(config_settings[store_key], list)
    val_is_str = isinstance(value, str)
    val_is_list = isinstance(value, list)
    if not key_exists:
        config_settings[store_key] = value
    elif key_exists_as_list and val_is_list:
        config_settings[store_key].extend(value)
    elif key_exists_as_list and val_is_str:
        config_settings[store_key].append(value)
    elif key_exists_as_str and val_is_str:
        msg = f"{key} already present in config and may not be overridden"
        raise RuntimeError(msg)
    else:
        msg = f"Type mismatch: cannot set {store_key} ({type(config_settings[store_key])}) to `{value}` ({type(value)})"
        raise RuntimeError(msg)


def build_sdist(sdist_directory: str, config_settings: Optional[dict[str, Union[list[str], str]]] = None) -> str:
    """Build a source distribution using the DuckDB submodule.

    This function extracts the DuckDB version from either the git submodule and saves it
    to a version file before building the sdist with scikit-build-core. If _FORCED_PEP440_VERSION
    was set then we first create a tag on the submodule.

    Args:
        sdist_directory: Directory where the sdist will be created.
        config_settings: Optional build configuration settings.

    Returns:
        The filename of the created sdist.

    Raises:
        RuntimeError: If not in a git repository or DuckDB submodule issues.
    """
    if not _in_git_repository():
        msg = "Not in a git repository, can't create an sdist"
        raise RuntimeError(msg)
    submodule_path = _duckdb_submodule_path()
    if _FORCED_PEP440_VERSION is not None:
        duckdb_version = pep440_to_git_tag(strip_post_from_version(_FORCED_PEP440_VERSION))
    else:
        duckdb_version = get_git_describe(repo_path=submodule_path, since_minor=MAIN_BRANCH_VERSIONING)
    _write_duckdb_long_version(duckdb_version)
    return skbuild_build_sdist(sdist_directory, config_settings=config_settings)


def build_wheel(
    wheel_directory: str,
    config_settings: Optional[dict[str, Union[list[str], str]]] = None,
    metadata_directory: Optional[str] = None,
) -> str:
    """Build a wheel from either git submodule or extracted sdist sources.

    This function builds a wheel using scikit-build-core, handling two scenarios:
    1. In a git repository: builds directly from the DuckDB submodule
    2. In an sdist: reads the saved DuckDB version and passes it to CMake

    Args:
        wheel_directory: Directory where the wheel will be created.
        config_settings: Optional build configuration settings.
        metadata_directory: Optional directory for metadata preparation.

    Returns:
        The filename of the created wheel.

    Raises:
        RuntimeError: If not in a git repository or sdist environment.
    """
    # First figure out the duckdb version we should use
    duckdb_version = None
    if not _in_git_repository():
        if not _in_sdist():
            msg = "Not in a git repository nor in an sdist, can't build a wheel"
            raise RuntimeError(msg)
        _log("Building duckdb wheel from sdist. Reading duckdb version from file.")
        config_settings = config_settings or {}
        duckdb_version = _read_duckdb_long_version()
    elif _FORCED_PEP440_VERSION is not None:
        duckdb_version = pep440_to_git_tag(strip_post_from_version(_FORCED_PEP440_VERSION))

    # We add the found version to the OVERRIDE_GIT_DESCRIBE cmake var
    if duckdb_version is not None:
        _skbuild_config_add(_SKBUILD_CMAKE_OVERRIDE_GIT_DESCRIBE, duckdb_version, config_settings)
        _log(f"{_SKBUILD_CMAKE_OVERRIDE_GIT_DESCRIBE} set to {duckdb_version}")
    else:
        _log("No explicit DuckDB submodule version provided. Letting CMake figure it out.")

    return skbuild_build_wheel(wheel_directory, config_settings=config_settings, metadata_directory=metadata_directory)


__all__ = [
    "build_editable",
    "build_sdist",
    "build_wheel",
    "get_requires_for_build_editable",
    "get_requires_for_build_sdist",
    "get_requires_for_build_wheel",
    "prepare_metadata_for_build_editable",
    "prepare_metadata_for_build_wheel",
]



import os

# list all include directories
include_directories = [
    os.path.sep.join(x.split('/'))
    for x in ['extension/tpcds/include', 'extension/tpcds/dsdgen/include', 'extension/tpcds/dsdgen/include/dsdgen-c']
]
# source files
source_files = [os.path.sep.join(x.split('/')) for x in ['extension/tpcds/tpcds_extension.cpp']]
source_files += [
    os.path.sep.join(x.split('/'))
    for x in [
        'extension/tpcds/dsdgen/dsdgen.cpp',
        'extension/tpcds/dsdgen/append_info-c.cpp',
        'extension/tpcds/dsdgen/dsdgen_helpers.cpp',
    ]
]
source_files += [
    os.path.sep.join(x.split('/'))
    for x in [
        'extension/tpcds/dsdgen/dsdgen-c/skip_days.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/address.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/build_support.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/date.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/dbgen_version.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/decimal.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/dist.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/error_msg.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/genrand.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/join.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/list.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/load.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/misc.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/nulls.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/parallel.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/permute.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/pricing.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/r_params.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/release.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/scaling.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/scd.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/sparse.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/StringBuffer.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/tdef_functions.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/tdefs.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/text.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_call_center.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_catalog_page.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_catalog_returns.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_catalog_sales.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_customer.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_customer_address.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_customer_demographics.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_datetbl.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_household_demographics.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_income_band.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_inventory.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_item.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_promotion.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_reason.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_ship_mode.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_store.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_store_returns.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_store_sales.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_timetbl.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_warehouse.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_web_page.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_web_returns.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_web_sales.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/w_web_site.cpp',
        'extension/tpcds/dsdgen/dsdgen-c/init.cpp',
    ]
]



import os
import platform

# list all include directories
include_directories = [
    os.path.sep.join(x.split('/')) for x in ['extension/jemalloc/include', 'extension/jemalloc/jemalloc/include']
]
# source files
source_files = [
    os.path.sep.join(x.split('/'))
    for x in [
        'extension/jemalloc/jemalloc_extension.cpp',
        'extension/jemalloc/jemalloc/src/jemalloc.c',
        'extension/jemalloc/jemalloc/src/arena.c',
        'extension/jemalloc/jemalloc/src/background_thread.c',
        'extension/jemalloc/jemalloc/src/base.c',
        'extension/jemalloc/jemalloc/src/batcher.c',
        'extension/jemalloc/jemalloc/src/bin.c',
        'extension/jemalloc/jemalloc/src/bin_info.c',
        'extension/jemalloc/jemalloc/src/bitmap.c',
        'extension/jemalloc/jemalloc/src/buf_writer.c',
        'extension/jemalloc/jemalloc/src/cache_bin.c',
        'extension/jemalloc/jemalloc/src/ckh.c',
        'extension/jemalloc/jemalloc/src/counter.c',
        'extension/jemalloc/jemalloc/src/ctl.c',
        'extension/jemalloc/jemalloc/src/decay.c',
        'extension/jemalloc/jemalloc/src/div.c',
        'extension/jemalloc/jemalloc/src/ecache.c',
        'extension/jemalloc/jemalloc/src/edata.c',
        'extension/jemalloc/jemalloc/src/edata_cache.c',
        'extension/jemalloc/jemalloc/src/ehooks.c',
        'extension/jemalloc/jemalloc/src/emap.c',
        'extension/jemalloc/jemalloc/src/eset.c',
        'extension/jemalloc/jemalloc/src/exp_grow.c',
        'extension/jemalloc/jemalloc/src/extent.c',
        'extension/jemalloc/jemalloc/src/extent_dss.c',
        'extension/jemalloc/jemalloc/src/extent_mmap.c',
        'extension/jemalloc/jemalloc/src/fxp.c',
        'extension/jemalloc/jemalloc/src/san.c',
        'extension/jemalloc/jemalloc/src/san_bump.c',
        'extension/jemalloc/jemalloc/src/hook.c',
        'extension/jemalloc/jemalloc/src/hpa.c',
        'extension/jemalloc/jemalloc/src/hpa_hooks.c',
        'extension/jemalloc/jemalloc/src/hpdata.c',
        'extension/jemalloc/jemalloc/src/inspect.c',
        'extension/jemalloc/jemalloc/src/large.c',
        'extension/jemalloc/jemalloc/src/log.c',
        'extension/jemalloc/jemalloc/src/malloc_io.c',
        'extension/jemalloc/jemalloc/src/mutex.c',
        'extension/jemalloc/jemalloc/src/nstime.c',
        'extension/jemalloc/jemalloc/src/pa.c',
        'extension/jemalloc/jemalloc/src/pa_extra.c',
        'extension/jemalloc/jemalloc/src/pai.c',
        'extension/jemalloc/jemalloc/src/pac.c',
        'extension/jemalloc/jemalloc/src/pages.c',
        'extension/jemalloc/jemalloc/src/peak_event.c',
        'extension/jemalloc/jemalloc/src/prof.c',
        'extension/jemalloc/jemalloc/src/prof_data.c',
        'extension/jemalloc/jemalloc/src/prof_log.c',
        'extension/jemalloc/jemalloc/src/prof_recent.c',
        'extension/jemalloc/jemalloc/src/prof_stats.c',
        'extension/jemalloc/jemalloc/src/prof_sys.c',
        'extension/jemalloc/jemalloc/src/psset.c',
        'extension/jemalloc/jemalloc/src/rtree.c',
        'extension/jemalloc/jemalloc/src/safety_check.c',
        'extension/jemalloc/jemalloc/src/sc.c',
        'extension/jemalloc/jemalloc/src/sec.c',
        'extension/jemalloc/jemalloc/src/stats.c',
        'extension/jemalloc/jemalloc/src/sz.c',
        'extension/jemalloc/jemalloc/src/tcache.c',
        'extension/jemalloc/jemalloc/src/test_hooks.c',
        'extension/jemalloc/jemalloc/src/thread_event.c',
        'extension/jemalloc/jemalloc/src/ticker.c',
        'extension/jemalloc/jemalloc/src/tsd.c',
        'extension/jemalloc/jemalloc/src/util.c',
        'extension/jemalloc/jemalloc/src/witness.c',
        'extension/jemalloc/jemalloc/src/zone.c',
    ]
]



import os

prefix = os.path.join('extension', 'core_functions')


def list_files_recursive(rootdir, suffix):
    file_list = []
    for root, _, files in os.walk(rootdir):
        file_list += [os.path.join(root, f) for f in files if f.endswith(suffix)]
    return file_list


include_directories = [os.path.join(prefix, x) for x in ['include']]
source_files = list_files_recursive(prefix, '.cpp')



"""!!HERE BE DRAGONS!! Use this script with care!

PyPI package cleanup tool. This script will:
* Never remove a stable version (including a post release version)
* Remove all release candidates for versions that have stable releases
* Remove all dev releases for versions that have stable releases
* Keep the configured amount of dev releases per version, and remove older dev releases
"""

import argparse
import contextlib
import heapq
import logging
import os
import re
import sys
import time
from collections import defaultdict
from collections.abc import Generator
from enum import Enum
from html.parser import HTMLParser
from typing import Optional, Union
from urllib.parse import urlparse

import pyotp
import requests
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from urllib3 import Retry

_PYPI_URL_PROD = "https://pypi.org/"
_PYPI_URL_TEST = "https://test.pypi.org/"
_DEFAULT_MAX_NIGHTLIES = 2
_LOGIN_RETRY_ATTEMPTS = 3
_LOGIN_RETRY_DELAY = 5


def create_argument_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser."""

    def max_nightlies_type(value: int) -> int:
        """Validate that --max-nightlies is set to a positive integer."""
        if int(value) < 0:
            msg = f"max-nightlies must be a positive integer, got {int(value)}"
            raise ValueError(msg)
        return int(value)

    parser = argparse.ArgumentParser(
        description="""
PyPI cleanup script for removing development versions.

!!HERE BE DRAGONS!! Use this script with care!

This script will:
* Never remove a stable version (including a post release version)
* Remove all release candidates for versions that have stable releases
* Remove all dev releases for versions that have stable releases
* Keep the configured amount of dev releases per version, and remove older dev releases
        """,
        epilog="Environment variables required (unless --dry-run): PYPI_CLEANUP_PASSWORD, PYPI_CLEANUP_OTP",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    loglevel_group = parser.add_mutually_exclusive_group(required=False)
    loglevel_group.add_argument(
        "-d",
        "--debug",
        help="Show debug logs",
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        default=logging.WARNING,
    )
    loglevel_group.add_argument(
        "-v",
        "--verbose",
        help="Show info logs",
        dest="loglevel",
        action="store_const",
        const=logging.INFO,
    )

    host_group = parser.add_mutually_exclusive_group(required=True)
    host_group.add_argument(
        "--prod", help="Use production PyPI (pypi.org)", dest="pypi_url", action="store_const", const=_PYPI_URL_PROD
    )
    host_group.add_argument(
        "--test", help="Use test PyPI (test.pypi.org)", dest="pypi_url", action="store_const", const=_PYPI_URL_TEST
    )

    parser.add_argument(
        "-m",
        "--max-nightlies",
        type=max_nightlies_type,
        default=_DEFAULT_MAX_NIGHTLIES,
        help=f"Max number of nightlies of unreleased versions (default={_DEFAULT_MAX_NIGHTLIES})",
    )

    subparsers = parser.add_subparsers(title="Subcommands")

    # Add the "list" subcommand
    parser_list = subparsers.add_parser("list", help="List all packages available for deletion")
    parser_list.set_defaults(func=lambda args: _run(CleanMode.LIST_ONLY, args))
    # Add the "delete" subcommand
    parser_delete = subparsers.add_parser(
        "delete", help="Delete packages that match the given criteria (use with care!"
    )
    parser_delete.add_argument(
        "-u", "--username", type=validate_username, help="PyPI username (required)", required=True
    )
    parser_delete.set_defaults(func=lambda args: _run(CleanMode.DELETE, args))

    return parser


class PyPICleanupError(Exception):
    """Base exception for PyPI cleanup operations."""


class AuthenticationError(PyPICleanupError):
    """Raised when authentication fails."""


class ValidationError(PyPICleanupError):
    """Raised when input validation fails."""


def setup_logging(level: int = logging.INFO) -> None:
    """Configure logging with appropriate level and format."""
    logging.basicConfig(level=level, format="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")


def validate_username(value: str) -> str:
    """Validate and sanitize username input."""
    if not value or not value.strip():
        msg = "Username cannot be empty"
        raise argparse.ArgumentTypeError(msg)

    username = value.strip()
    if len(username) > 100:  # Reasonable limit
        msg = "Username too long (max 100 characters)"
        raise argparse.ArgumentTypeError(msg)

    # Basic validation - PyPI usernames are alphanumeric with limited special chars
    if not re.match(r"^[a-zA-Z0-9][a-zA-Z0-9._-]*[a-zA-Z0-9]$|^[a-zA-Z0-9]$", username):
        msg = "Invalid username format"
        raise argparse.ArgumentTypeError(msg)

    return username


@contextlib.contextmanager
def session_with_retries() -> Generator[Session, None, None]:
    """Create a requests session with retry strategy for ephemeral errors."""
    with requests.Session() as session:
        retry_strategy = Retry(
            allowed_methods=["GET", "POST"],
            total=None,  # disable to make the below take effect
            redirect=10,  # Don't follow more than 10 redirects in a row
            connect=3,  # try 3 times before giving up on connection errors
            read=3,  # try 3 times before giving up on read errors
            status=3,  # try 3 times before giving up on status errors (see forcelist below)
            status_forcelist=[429, *list(range(500, 512))],
            other=0,  # whatever else may cause an error should break
            backoff_factor=0.1,  # [0.0s, 0.2s, 0.4s]
            raise_on_redirect=True,  # raise exception when redirect error retries are exhausted
            raise_on_status=True,  # raise exception when status error retries are exhausted
            respect_retry_after_header=True,  # respect Retry-After headers
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("https://", adapter)
        yield session


def load_credentials() -> tuple[Optional[str], Optional[str]]:
    """Load credentials from environment variables."""
    password = os.getenv("PYPI_CLEANUP_PASSWORD")
    otp = os.getenv("PYPI_CLEANUP_OTP")

    if not password:
        msg = "PYPI_CLEANUP_PASSWORD environment variable is required when not in dry-run mode"
        raise ValidationError(msg)
    if not otp:
        msg = "PYPI_CLEANUP_OTP environment variable is required when not in dry-run mode"
        raise ValidationError(msg)

    return password, otp


class CsrfParser(HTMLParser):
    """HTML parser to extract CSRF tokens from PyPI forms.

    Based on pypi-cleanup package (https://github.com/arcivanov/pypi-cleanup/tree/master)
    """

    def __init__(self, target: str) -> None:  # noqa: D107
        super().__init__()
        self._target = target
        self.csrf = None  # Result value from all forms on page
        self._in_form = False  # Currently parsing a form with an action we're interested in

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Union[str, None]]]) -> None:  # noqa: D102
        if not self.csrf:
            if tag == "form":
                attrs = dict(attrs)
                action = attrs.get("action")  # Might be None.
                if action and (action == self._target or action.startswith(self._target)):
                    self._in_form = True
            elif self._in_form and tag == "input":
                attrs = dict(attrs)
                if attrs.get("name") == "csrf_token" and not self.csrf:
                    self.csrf = attrs["value"]

    def handle_endtag(self, tag: str) -> None:  # noqa: D102
        if tag == "form" and self._in_form:
            self._in_form = False


class CleanMode(Enum):
    """Supported clean-up modes."""

    LIST_ONLY = 1
    DELETE = 2


class PyPICleanup:
    """Main class for performing PyPI package cleanup operations."""

    def __init__(  # noqa: D107
        self,
        index_url: str,
        mode: CleanMode,
        max_dev_releases: int = _DEFAULT_MAX_NIGHTLIES,
        username: Optional[str] = None,
        password: Optional[str] = None,
        otp: Optional[str] = None,
    ) -> None:
        parsed_url = urlparse(index_url)
        self._index_url = parsed_url.geturl().rstrip("/")
        self._index_host = parsed_url.hostname
        self._mode = mode
        self._max_dev_releases = max_dev_releases
        self._username = username
        self._password = password
        self._otp = otp
        self._package = "duckdb"
        self._dev_version_pattern = re.compile(r"^(?P<version>\d+\.\d+\.\d+)\.dev(?P<dev_id>\d+)$")
        self._rc_version_pattern = re.compile(r"^(?P<version>\d+\.\d+\.\d+)\.rc\d+$")
        self._stable_version_pattern = re.compile(r"^\d+\.\d+\.\d+(\.post\d+)?$")

    def run(self) -> int:
        """Execute the cleanup process.

        Returns:
            int: Exit code (0 for success, non-zero for failure)
        """
        if self._mode == CleanMode.DELETE:
            logging.warning("NOT A DRILL: WILL DELETE PACKAGES")
        elif self._mode == CleanMode.LIST_ONLY:
            logging.debug("Running in DRY RUN mode, nothing will be deleted")
        else:
            msg = "Unexpected mode"
            raise RuntimeError(msg)

        logging.info(f"Max development releases to keep per unreleased version: {self._max_dev_releases}")

        try:
            with session_with_retries() as http_session:
                return self._execute_cleanup(http_session)
        except PyPICleanupError:
            logging.exception("Cleanup failed")
            return 1
        except Exception as e:
            logging.error(f"Unexpected error: {e}", exc_info=True)
            return 1

    def _execute_cleanup(self, http_session: Session) -> int:
        """Execute the main cleanup logic."""
        # Get released versions
        versions = self._fetch_released_versions(http_session)
        if not versions:
            logging.info(f"No releases found for {self._package}")
            return 0

        # Determine and report versions to delete
        versions_to_delete = self._determine_versions_to_delete(versions)
        if len(versions_to_delete) > 0:
            print(f"Found the following stale releases on {self._index_host}:")
            for version in sorted(versions_to_delete):
                print(f"- {version}")
        else:
            print(f"No stale releases found on {self._index_host}")
            return 0

        if self._mode != CleanMode.DELETE:
            logging.info("Dry run complete - no packages were deleted")
            return 0

        logging.warning(f"Will try to delete {len(versions_to_delete)} releases from {self._index_host}")

        # Perform authentication and deletion
        self._authenticate(http_session)
        self._delete_versions(http_session, versions_to_delete)

        logging.info(f"Successfully cleaned up {len(versions_to_delete)} development versions")
        return 0

    def _fetch_released_versions(self, http_session: Session) -> set[str]:
        """Fetch package release information from PyPI API."""
        logging.debug(f"Fetching package information for '{self._package}'")

        try:
            req = http_session.get(f"{self._index_url}/pypi/{self._package}/json")
            req.raise_for_status()
        except RequestException as e:
            msg = f"Failed to fetch package information for '{self._package}': {e}"
            raise PyPICleanupError(msg) from e

        data = req.json()
        versions = {v for v, files in data["releases"].items() if len(files) > 0}
        logging.debug(f"Found {len(versions)} releases with files")
        return versions

    def _is_stable_release_version(self, version: str) -> bool:
        """Determine whether a version string denotes a stable release."""
        return self._stable_version_pattern.match(version) is not None

    def _is_rc_version(self, version: str) -> bool:
        """Determine whether a version string denotes a stable release."""
        return self._rc_version_pattern.match(version) is not None

    def _is_dev_version(self, version: str) -> bool:
        """Determine whether a version string denotes a dev release."""
        return self._dev_version_pattern.match(version) is not None

    def _parse_rc_version(self, version: str) -> str:
        """Parse a rc version string to determine the base version."""
        match = self._rc_version_pattern.match(version)
        if not match:
            msg = f"Invalid rc version '{version}'"
            raise PyPICleanupError(msg)
        return match.group("version") if match else None

    def _parse_dev_version(self, version: str) -> tuple[str, int]:
        """Parse a dev version string to determine the base version and dev version id."""
        match = self._dev_version_pattern.match(version)
        if not match:
            msg = f"Invalid dev version '{version}'"
            raise PyPICleanupError(msg)
        return match.group("version"), int(match.group("dev_id"))

    def _determine_versions_to_delete(self, versions: set[str]) -> set[str]:
        """Determine which package versions should be deleted."""
        logging.debug("Analyzing versions to determine cleanup candidates")

        # Get all stable, rc and dev versions
        stable_versions = {v for v in versions if self._is_stable_release_version(v)}
        rc_versions = {v for v in versions if self._is_rc_version(v)}
        rc_base_versions = {self._parse_rc_version(v) for v in versions if self._is_rc_version(v)}
        dev_versions = {v for v in versions if self._is_dev_version(v)}

        # Set of all rc releases of versions that have a stable release
        rcs_of_stable = {v for v in rc_versions if self._parse_rc_version(v) in stable_versions}
        # Set of all dev releases of versions that have a stable or rc release
        devs_of_stable = {v for v in dev_versions if self._parse_dev_version(v)[0] in stable_versions}
        devs_of_rc = {v for v in dev_versions if self._parse_dev_version(v)[0] in rc_base_versions}
        # Set of orphan dev versions
        orphan_devs = dev_versions.difference(devs_of_stable).difference(devs_of_rc)

        # Construct list of orphan dev
        orphan_devs_per_version = defaultdict(list)
        # 1. put all dev keep candidates on a max heap indexed by negative dev id (i.e. dev10 -> -10)
        for version in orphan_devs:
            base_version, dev_id = self._parse_dev_version(version)
            heapq.heappush(orphan_devs_per_version[base_version], (-dev_id, version))
        # 2. remove the amount of latest dev releases we want to keep
        for version_list in orphan_devs_per_version.values():
            for _ in range(min(self._max_dev_releases, len(version_list))):
                heapq.heappop(version_list)
        # 3. Result: set of outdated dev versions
        devs_outdated = {v for version_list in orphan_devs_per_version.values() for _, v in version_list}

        # Construct final deletion set
        versions_to_delete = set()
        if rcs_of_stable:
            versions_to_delete.update(rcs_of_stable)
            logging.info(f"Found {len(rcs_of_stable)} release candidates that have stable releases")
        if devs_of_stable:
            versions_to_delete.update(devs_of_stable)
            logging.info(f"Found {len(devs_of_stable)} dev releases that have stable releases")
        if devs_of_rc:
            versions_to_delete.update(devs_of_rc)
            logging.info(f"Found {len(devs_of_rc)} dev releases that have release candidates")
        if devs_outdated:
            versions_to_delete.update(devs_outdated)
            logging.info(f"Found {len(devs_outdated)} dev releases that are outdated")

        # Final safety checks
        if versions_to_delete == versions:
            msg = (
                f"Safety check failed: cleanup would delete ALL versions of '{self._package}'. "
                "This would make the package permanently inaccessible. Aborting."
            )
            raise PyPICleanupError(msg)
        if len(versions_to_delete.intersection(stable_versions)) > 0:
            msg = (
                f"Safety check failed: cleanup would delete one or more stable versions of '{self._package}'. "
                f"A regexp might be broken? (would delete {versions_to_delete.intersection(stable_versions)})"
            )
            raise PyPICleanupError(msg)
        unknown_versions = versions.difference(stable_versions).difference(rc_versions).difference(dev_versions)
        if unknown_versions:
            logging.warning(f"Found version string(s) in an unsupported format: {unknown_versions}")

        return versions_to_delete

    def _authenticate(self, http_session: Session) -> None:
        """Authenticate with PyPI."""
        if not self._username or not self._password:
            msg = "Username and password are required for authentication"
            raise AuthenticationError(msg)

        logging.info(f"Authenticating user '{self._username}' with PyPI")

        try:
            # Attempt login
            login_response = self._perform_login(http_session)

            # Handle two-factor authentication if required
            if login_response.url.startswith(f"{self._index_url}/account/two-factor/"):
                logging.debug("Two-factor authentication required")
                self._handle_two_factor_auth(http_session, login_response)

            logging.info("Authentication successful")

        except RequestException as e:
            msg = f"Network error during authentication: {e}"
            raise AuthenticationError(msg) from e

    def _get_csrf_token(self, http_session: Session, form_action: str) -> str:
        """Extract CSRF token from a form page."""
        resp = http_session.get(f"{self._index_url}{form_action}")
        resp.raise_for_status()
        parser = CsrfParser(form_action)
        parser.feed(resp.text)
        if not parser.csrf:
            msg = f"No CSRF token found in {form_action}"
            raise AuthenticationError(msg)
        return parser.csrf

    def _perform_login(self, http_session: Session) -> requests.Response:
        """Perform the initial login with username/password."""
        # Get login form and CSRF token
        csrf_token = self._get_csrf_token(http_session, "/account/login/")

        login_data = {"csrf_token": csrf_token, "username": self._username, "password": self._password}

        response = http_session.post(
            f"{self._index_url}/account/login/",
            data=login_data,
            headers={"referer": f"{self._index_url}/account/login/"},
        )
        response.raise_for_status()

        # Check if login failed (redirected back to login page)
        if response.url == f"{self._index_url}/account/login/":
            msg = f"Login failed for user '{self._username}' - check credentials"
            raise AuthenticationError(msg)

        return response

    def _handle_two_factor_auth(self, http_session: Session, response: requests.Response) -> None:
        """Handle two-factor authentication."""
        if not self._otp:
            msg = "Two-factor authentication required but no OTP secret provided"
            raise AuthenticationError(msg)

        two_factor_url = response.url
        form_action = two_factor_url[len(self._index_url) :]
        csrf_token = self._get_csrf_token(http_session, form_action)

        # Try authentication with retries
        for attempt in range(_LOGIN_RETRY_ATTEMPTS):
            try:
                auth_code = pyotp.TOTP(self._otp).now()
                logging.debug(f"Attempting 2FA with code (attempt {attempt + 1}/{_LOGIN_RETRY_ATTEMPTS})")

                auth_response = http_session.post(
                    two_factor_url,
                    data={"csrf_token": csrf_token, "method": "totp", "totp_value": auth_code},
                    headers={"referer": two_factor_url},
                )
                auth_response.raise_for_status()

                # Check if 2FA succeeded (redirected away from 2FA page)
                if auth_response.url != two_factor_url:
                    logging.debug("Two-factor authentication successful")
                    return

                if attempt < _LOGIN_RETRY_ATTEMPTS - 1:
                    logging.debug(f"2FA code rejected, retrying in {_LOGIN_RETRY_DELAY} seconds...")
                    time.sleep(_LOGIN_RETRY_DELAY)

            except RequestException as e:
                if attempt == _LOGIN_RETRY_ATTEMPTS - 1:
                    msg = f"Network error during 2FA: {e}"
                    raise AuthenticationError(msg) from e
                logging.debug(f"Network error during 2FA attempt {attempt + 1}, retrying...")
                time.sleep(_LOGIN_RETRY_DELAY)

        msg = "Two-factor authentication failed after all attempts"
        raise AuthenticationError(msg)

    def _delete_versions(self, http_session: Session, versions_to_delete: set[str]) -> None:
        """Delete the specified package versions."""
        logging.info(f"Starting deletion of {len(versions_to_delete)} development versions")

        failed_deletions = []
        for version in sorted(versions_to_delete):
            try:
                self._delete_single_version(http_session, version)
                logging.info(f"Successfully deleted {self._package} version {version}")
            except Exception:
                # Continue with other versions rather than failing completely
                logging.exception(f"Failed to delete version {version}")
                failed_deletions.append(version)

        if failed_deletions:
            msg = f"Failed to delete {len(failed_deletions)}/{len(versions_to_delete)} versions: {failed_deletions}"
            raise PyPICleanupError(msg)

    def _delete_single_version(self, http_session: Session, version: str) -> None:
        """Delete a single package version."""
        # Safety check
        if not self._is_dev_version(version) or self._is_rc_version(version):
            msg = f"Refusing to delete non-[dev|rc] version: {version}"
            raise PyPICleanupError(msg)

        logging.debug(f"Deleting {self._package} version {version}")

        # Get deletion form and CSRF token
        form_action = f"/manage/project/{self._package}/release/{version}/"
        form_url = f"{self._index_url}{form_action}"

        csrf_token = self._get_csrf_token(http_session, form_action)

        # Submit deletion request
        delete_response = http_session.post(
            form_url,
            data={
                "csrf_token": csrf_token,
                "confirm_delete_version": version,
            },
            headers={"referer": form_url},
        )
        delete_response.raise_for_status()


def _run(mode: CleanMode, args: argparse.Namespace) -> int:
    """Action called by the subcommands after arg parsing."""
    setup_logging(args.loglevel)
    try:
        if mode == CleanMode.DELETE:
            password, otp = load_credentials()
            cleanup = PyPICleanup(
                args.pypi_url, mode, args.max_nightlies, username=args.username, password=password, otp=otp
            )
        elif mode == CleanMode.LIST_ONLY:
            cleanup = PyPICleanup(args.pypi_url, mode, args.max_nightlies)
        else:
            print(f"Unknown mode {mode}. Did nothing.")
            return -1
        return cleanup.run()
    except ValidationError:
        logging.exception("Configuration error")
        return 2
    except KeyboardInterrupt:
        logging.info("Operation cancelled by user")
        return 130
    except Exception:
        logging.exception("Unexpected error")
        return 1


def main() -> int:
    """Main entry point for the script."""
    parser = create_argument_parser()
    args = parser.parse_args()
    # call the subcommand's func
    return args.func(args)


if __name__ == "__main__":
    sys.exit(main())



"""DuckDB Python versioning utilities. This will only work on Python >= 3.3 and on non-mobile platforms.

This module provides utilities for version management including:
- Version bumping (major, minor, patch, post)
- Git tag creation and management
- Version parsing and validation
"""

import pathlib
import re
import subprocess
from typing import Optional

VERSION_RE = re.compile(
    r"^(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)(?:rc(?P<rc>[0-9]+)|\.post(?P<post>[0-9]+))?$"
)


def parse_version(version: str) -> tuple[int, int, int, int, int]:
    """Parse a version string into its components.

    Args:
        version: Version string (e.g., "1.3.1", "1.3.2.rc3" or "1.3.1.post2")

    Returns:
        Tuple of (major, minor, patch, post, rc)

    Raises:
        ValueError: If version format is invalid
    """
    match = VERSION_RE.match(version)
    if not match:
        msg = f"Invalid version format: {version} (expected X.Y.Z, X.Y.Z.rcM or X.Y.Z.postN)"
        raise ValueError(msg)

    major, minor, patch, rc, post = match.groups()
    return int(major), int(minor), int(patch), int(post or 0), int(rc or 0)


def format_version(major: int, minor: int, patch: int, post: int = 0, rc: int = 0) -> str:
    """Format version components into a version string.

    Args:
        major: Major version number
        minor: Minor version number
        patch: Patch version number
        post: Post-release number
        rc: RC number

    Returns:
        Formatted version string
    """
    version = f"{major}.{minor}.{patch}"
    if post != 0 and rc != 0:
        msg = "post and rc are mutually exclusive"
        raise ValueError(msg)
    if post != 0:
        version += f".post{post}"
    if rc != 0:
        version += f"rc{rc}"
    return version


def git_tag_to_pep440(git_tag: str) -> str:
    """Convert git tag format to PEP440 format.

    Args:
        git_tag: Git tag (e.g., "v1.3.1", "v1.3.1-post1")

    Returns:
        PEP440 version string (e.g., "1.3.1", "1.3.1.post1")
    """
    # Remove 'v' prefix if present
    version = git_tag[1:] if git_tag.startswith("v") else git_tag

    if "-post" in version:
        assert "rc" not in version
        version = version.replace("-post", ".post")
    elif "-rc" in version:
        version = version.replace("-rc", "rc")

    return version


def pep440_to_git_tag(version: str) -> str:
    """Convert PEP440 version to git tag format.

    Args:
        version: PEP440 version string (e.g., "1.3.1.post1" or "1.3.1rc2")

    Returns:
        Git tag format (e.g., "v1.3.1-post1")
    """
    if ".post" in version:
        assert "rc" not in version
        version = version.replace(".post", "-post")
    elif "rc" in version:
        version = version.replace("rc", "-rc")

    return f"v{version}"


def get_current_version() -> Optional[str]:
    """Get the current version from git tags.

    Returns:
        Current version string or None if no tags exist
    """
    try:
        # Get the latest tag
        result = subprocess.run(["git", "describe", "--tags", "--abbrev=0"], capture_output=True, text=True, check=True)
        tag = result.stdout.strip()
        return git_tag_to_pep440(tag)
    except subprocess.CalledProcessError:
        return None


def create_git_tag(version: str, message: Optional[str] = None, repo_path: Optional[pathlib.Path] = None) -> None:
    """Create a git tag for the given version.

    Args:
        version: Version string (PEP440 format)
        message: Optional tag message
        repo_path: Optional path to git repository (defaults to current directory)

    Raises:
        subprocess.CalledProcessError: If git command fails
    """
    tag_name = pep440_to_git_tag(version)

    cmd = ["git", "tag"]
    if message:
        cmd.extend(["-a", tag_name, "-m", message])
    else:
        cmd.append(tag_name)

    # If a repository path is provided, use it as the working directory
    cwd = repo_path if repo_path is not None else None
    subprocess.run(cmd, check=True, cwd=cwd)


def strip_post_from_version(version: str) -> str:
    """Removing post-release suffixes from the given version.

    DuckDB doesn't allow post-release versions, so .post* suffixes are stripped.
    """
    return re.sub(r"[\.-]post[0-9]+", "", version)


def get_git_describe(
    repo_path: Optional[pathlib.Path] = None,
    since_major: bool = False,  # noqa: FBT001
    since_minor: bool = False,  # noqa: FBT001
) -> Optional[str]:
    """Get git describe output for version determination.

    Returns:
        Git describe output or None if no tags exist
    """
    cwd = repo_path if repo_path is not None else None
    pattern = "v*.*.*"
    if since_major:
        pattern = "v*.0.0"
    elif since_minor:
        pattern = "v*.*.0"
    try:
        result = subprocess.run(
            ["git", "describe", "--tags", "--long", "--match", pattern],
            capture_output=True,
            text=True,
            check=True,
            cwd=cwd,
        )
        result.check_returncode()
        return result.stdout.strip()
    except FileNotFoundError as e:
        msg = "git executable can't be found"
        raise RuntimeError(msg) from e
